"""Provide class to read informations on vector data from file path or object path

The module contains the following class :

    - 'Vector' - Data Vector

"""

# -- IMPORTS --

# standard library
import os
import tempfile

# 3rd party
from osgeo import ogr

# package
from rok4.storage import copy, get_osgeo_path

# -- GLOBALS --

# Enable GDAL/OGR exceptions
ogr.UseExceptions()


class Vector:
    """A data vector

    Attributes:
        path (str): path to the file/object
        bbox (Tuple[float, float, float, float]): bounding rectange in the data projection
        layers (List[Tuple[str, int, List[Tuple[str, str]]]]) : Vector layers with their name, their number of objects and their attributes
    """

    @classmethod
    def from_file(cls, path: str, **kwargs) -> "Vector":
        """Constructor method of a Vector from a file (Shapefile, Geopackage, CSV and GeoJSON)

        Args:
            path (str): path to the file/object
            **csv (Dict[str : str]) : dictionnary of CSV parameters :
                -srs (str) ("EPSG:2154" if not provided) : spatial reference system of the geometry
                -column_x (str) ("x" if not provided) : field of the x coordinate
                -column_y (str) ("y" if not provided) : field of the y coordinate
                -column_wkt (str) (None if not provided) : field of the WKT of the geometry if WKT use to define coordinate

        Examples:

            from rok4.vector import Vector

            try:
                vector = Vector.from_file("file://tests/fixtures/ARRONDISSEMENT.shp")
                vector_csv1 = Vector.from_file("file://tests/fixtures/vector.csv" , csv={"delimiter":";", "column_x":"x", "column_y":"y"})
                vector_csv2 = Vector.from_file("file://tests/fixtures/vector2.csv" , csv={"delimiter":";", "column_wkt":"WKT"})

            except Exception as e:
                print(f"Vector creation raises an exception: {exc}")

        Raises:
            MissingEnvironmentError: Missing object storage informations
            StorageError: Storage read issue
            Exception: Wrong column
            Exception: Wrong data in column
            Exception: Wrong format of file
            Exception: Wrong data in the file

        """

        self = cls()

        self.path = path

        path_split = path.split("/")

        if path_split[0] == "ceph:" or path.endswith(".csv"):
            if path.endswith(".shp"):
                with tempfile.TemporaryDirectory() as tmp:
                    tmp_path = tmp + "/" + path_split[-1][:-4]

                    copy(path, "file://" + tmp_path + ".shp")
                    copy(path[:-4] + ".shx", "file://" + tmp_path + ".shx")
                    copy(path[:-4] + ".cpg", "file://" + tmp_path + ".cpg")
                    copy(path[:-4] + ".dbf", "file://" + tmp_path + ".dbf")
                    copy(path[:-4] + ".prj", "file://" + tmp_path + ".prj")

                    dataSource = ogr.Open(tmp_path + ".shp", 0)

            elif path.endswith(".gpkg"):
                with tempfile.TemporaryDirectory() as tmp:
                    tmp_path = tmp + "/" + path_split[-1][:-5]

                    copy(path, "file://" + tmp_path + ".gpkg")

                    dataSource = ogr.Open(tmp_path + ".gpkg", 0)

            elif path.endswith(".geojson"):
                with tempfile.TemporaryDirectory() as tmp:
                    tmp_path = tmp + "/" + path_split[-1][:-8]

                    copy(path, "file://" + tmp_path + ".geojson")

                    dataSource = ogr.Open(tmp_path + ".geojson", 0)

            elif path.endswith(".csv"):
                # Récupération des informations optionnelles
                if "csv" in kwargs:
                    csv = kwargs["csv"]
                else:
                    csv = {}

                if "srs" in csv and csv["srs"] is not None:
                    srs = csv["srs"]
                else:
                    srs = "EPSG:2154"

                if "column_x" in csv and csv["column_x"] is not None:
                    column_x = csv["column_x"]
                else:
                    column_x = "x"

                if "column_y" in csv and csv["column_y"] is not None:
                    column_y = csv["column_y"]
                else:
                    column_y = "y"

                if "column_wkt" in csv:
                    column_wkt = csv["column_wkt"]
                else:
                    column_wkt = None

                with tempfile.TemporaryDirectory() as tmp:
                    tmp_path = tmp + "/" + path_split[-1][:-4]
                    name_fich = path_split[-1][:-4]

                    copy(path, "file://" + tmp_path + ".csv")

                    with tempfile.NamedTemporaryFile(
                        mode="w", suffix=".vrt", dir=tmp, delete=False
                    ) as tmp2:
                        vrt_file = "<OGRVRTDataSource>\n"
                        vrt_file += '<OGRVRTLayer name="' + name_fich + '">\n'
                        vrt_file += "<SrcDataSource>" + tmp_path + ".csv</SrcDataSource>\n"
                        vrt_file += "<SrcLayer>" + name_fich + "</SrcLayer>\n"
                        vrt_file += "<LayerSRS>" + srs + "</LayerSRS>\n"
                        if column_wkt is None:
                            vrt_file += (
                                '<GeometryField encoding="PointFromColumns" x="'
                                + column_x
                                + '" y="'
                                + column_y
                                + '"/>\n'
                            )
                        else:
                            vrt_file += (
                                '<GeometryField encoding="WKT" field="' + column_wkt + '"/>\n'
                            )
                        vrt_file += "</OGRVRTLayer>\n"
                        vrt_file += "</OGRVRTDataSource>"
                        tmp2.write(vrt_file)
                    dataSourceVRT = ogr.Open(tmp2.name, 0)
                    os.remove(tmp2.name)
                    dataSource = ogr.GetDriverByName("ESRI Shapefile").CopyDataSource(
                        dataSourceVRT, tmp_path + "shp"
                    )

            else:
                raise Exception("This format of file cannot be loaded")

        else:
            dataSource = ogr.Open(get_osgeo_path(path), 0)

        multipolygon = ogr.Geometry(ogr.wkbGeometryCollection)
        try:
            layer = dataSource.GetLayer()
        except AttributeError:
            raise Exception(f"The content of {self.path} cannot be read")

        layers = []
        for i in range(dataSource.GetLayerCount()):
            layer = dataSource.GetLayer(i)
            name = layer.GetName()
            count = layer.GetFeatureCount()
            layerDefinition = layer.GetLayerDefn()
            attributes = []
            for j in range(layerDefinition.GetFieldCount()):
                fieldName = layerDefinition.GetFieldDefn(j).GetName()
                fieldTypeCode = layerDefinition.GetFieldDefn(j).GetType()
                fieldType = layerDefinition.GetFieldDefn(j).GetFieldTypeName(fieldTypeCode)
                attributes += [(fieldName, fieldType)]
            for feature in layer:
                geom = feature.GetGeometryRef()
                if geom is not None:
                    multipolygon.AddGeometry(geom)
            layers += [(name, count, attributes)]

        self.layers = layers
        self.bbox = multipolygon.GetEnvelope()

        return self

    @classmethod
    def from_parameters(cls, path: str, bbox: tuple, layers: list) -> "Vector":
        """Constructor method of a Vector from a parameters

        Args:
            path (str): path to the file/object
            bbox (Tuple[float, float, float, float]): bounding rectange in the data projection
            layers (List[Tuple[str, int, List[Tuple[str, str]]]]) : Vector layers with their name, their number of objects and their attributes

        Examples:

            try :
                vector = Vector.from_parameters("file://tests/fixtures/ARRONDISSEMENT.shp", (1,2,3,4), [('ARRONDISSEMENT', 14, [('ID', 'String'), ('NOM', 'String'), ('INSEE_ARR', 'String'), ('INSEE_DEP', 'String'), ('INSEE_REG', 'String'), ('ID_AUT_ADM', 'String'), ('DATE_CREAT', 'String'), ('DATE_MAJ', 'String'), ('DATE_APP', 'Date'), ('DATE_CONF', 'Date')])])

            except Exception as e:
                print(f"Vector creation raises an exception: {exc}")

        """

        self = cls()

        self.path = path
        self.bbox = bbox
        self.layers = layers

        return self



"""Describes unit tests for the rok4.raster module.

There is one test class for each tested functionnality.
See internal docstrings for more information.
Each variable prefixed by "m_" is a mock, or part of it.
"""

import copy
import json
import math
import random
from unittest import TestCase, mock
from unittest.mock import MagicMock, call, mock_open

import pytest

from rok4.enums import ColorFormat
from rok4.raster import Raster, RasterSet

# rok4.raster.Raster class tests


class TestRasterInit(TestCase):
    """rok4.raster.Raster default constructor."""

    def test_default(self):
        """Default property values."""
        raster = Raster()

        assert raster.bands is None
        assert (
            isinstance(raster.bbox, tuple)
            and len(raster.bbox) == 4
            and all(coordinate is None for coordinate in raster.bbox)
        )
        assert (
            isinstance(raster.dimensions, tuple)
            and len(raster.dimensions) == 2
            and all(dimension is None for dimension in raster.dimensions)
        )
        assert raster.format is None
        assert raster.mask is None
        assert raster.path is None


class TestRasterFromFile(TestCase):
    """rok4.raster.Raster.from_file(path) class constructor."""

    def setUp(self):
        self.source_image_path = "file:///home/user/image.tif"
        self.source_mask_path = "file:///home/user/image.msk"
        self.osgeo_image_path = "file:///home/user/image.tif"
        self.osgeo_mask_path = "file:///home/user/image.msk"
        self.bbox = (-5.4, 41.3, 9.8, 51.3)
        self.image_size = (1920, 1080)
        return super().setUp()

    def test_empty(self):
        """Constructor called without the expected path argument."""
        with pytest.raises(TypeError):
            Raster.from_file()

    @mock.patch("rok4.raster.exists", return_value=False)
    def test_image_not_found(self, m_exists):
        """Constructor called on a path matching no file or object."""
        with pytest.raises(Exception):
            Raster.from_file(self.source_image_path)
        m_exists.assert_called_once_with(self.source_image_path)

    @mock.patch("rok4.raster.get_osgeo_path")
    @mock.patch("rok4.raster.compute_format", return_value=ColorFormat.UINT8)
    @mock.patch("rok4.raster.gdal.Open")
    @mock.patch("rok4.raster.compute_bbox")
    @mock.patch("rok4.raster.exists", side_effect=[True, False])
    def test_image(self, m_exists, m_compute_bbox, m_gdal_open, m_compute_format, m_get_osgeo_path):
        """Constructor called nominally on an image without mask."""
        m_compute_bbox.return_value = self.bbox
        m_dataset_properties = {
            "RasterCount": 3,
            "RasterXSize": self.image_size[0],
            "RasterYSize": self.image_size[1],
        }
        m_gdal_open.return_value = type("", (object,), m_dataset_properties)
        m_get_osgeo_path.return_value = self.osgeo_image_path

        raster = Raster.from_file(self.source_image_path)

        m_exists.assert_has_calls([call(self.source_image_path), call(self.source_mask_path)])
        m_get_osgeo_path.assert_called_once_with(self.source_image_path)
        m_gdal_open.assert_called_once_with(self.osgeo_image_path)
        assert raster.path == self.source_image_path
        assert raster.mask is None
        m_compute_bbox.assert_called_once()
        assert (
            isinstance(raster.bbox, tuple)
            and len(raster.bbox) == 4
            and math.isclose(raster.bbox[0], self.bbox[0], rel_tol=1e-5)
            and math.isclose(raster.bbox[1], self.bbox[1], rel_tol=1e-5)
            and math.isclose(raster.bbox[2], self.bbox[2], rel_tol=1e-5)
            and math.isclose(raster.bbox[3], self.bbox[3], rel_tol=1e-5)
        )
        assert raster.bands == 3
        m_compute_format.assert_called_once()
        assert raster.format == ColorFormat.UINT8
        assert raster.dimensions == self.image_size

    @mock.patch("rok4.raster.get_osgeo_path")
    @mock.patch("rok4.raster.compute_format", return_value=ColorFormat.UINT8)
    @mock.patch("rok4.raster.gdal.IdentifyDriver")
    @mock.patch("rok4.raster.gdal.Open")
    @mock.patch("rok4.raster.compute_bbox")
    @mock.patch("rok4.raster.exists", side_effect=[True, True])
    def test_image_and_mask(
        self,
        m_exists,
        m_compute_bbox,
        m_gdal_open,
        m_identifydriver,
        m_compute_format,
        m_get_osgeo_path,
    ):
        """Constructor called nominally on an image with mask."""
        m_compute_bbox.return_value = self.bbox
        m_dataset_properties = {
            "RasterCount": 3,
            "RasterXSize": self.image_size[0],
            "RasterYSize": self.image_size[1],
        }
        m_gdal_open.return_value = type("", (object,), m_dataset_properties)
        m_get_osgeo_path.side_effect = [self.osgeo_image_path, self.osgeo_mask_path]
        m_identifydriver.return_value = type("", (object,), {"ShortName": "GTiff"})

        raster = Raster.from_file(self.source_image_path)

        m_exists.assert_has_calls([call(self.source_image_path), call(self.source_mask_path)])
        m_get_osgeo_path.assert_has_calls(
            [call(self.source_image_path), call(self.source_mask_path)]
        )
        m_identifydriver.assert_called_once_with(self.osgeo_mask_path)
        m_gdal_open.assert_called_once_with(self.osgeo_image_path)
        assert raster.path == self.source_image_path
        assert raster.mask == self.source_mask_path
        m_compute_bbox.assert_called_once()
        assert (
            isinstance(raster.bbox, tuple)
            and len(raster.bbox) == 4
            and math.isclose(raster.bbox[0], self.bbox[0], rel_tol=1e-5)
            and math.isclose(raster.bbox[1], self.bbox[1], rel_tol=1e-5)
            and math.isclose(raster.bbox[2], self.bbox[2], rel_tol=1e-5)
            and math.isclose(raster.bbox[3], self.bbox[3], rel_tol=1e-5)
        )
        assert raster.bands == 3
        m_compute_format.assert_called_once()
        assert raster.format == ColorFormat.UINT8
        assert raster.dimensions == self.image_size

    @mock.patch("rok4.raster.get_osgeo_path")
    @mock.patch("rok4.raster.gdal.Open", side_effect=RuntimeError)
    @mock.patch("rok4.raster.exists", side_effect=[True, False])
    def test_unsupported_image_format(self, m_exists, m_gdal_open, m_get_osgeo_path):
        """Test case : Constructor called on an unsupported image file or object."""
        m_get_osgeo_path.return_value = self.osgeo_image_path

        with pytest.raises(RuntimeError):
            Raster.from_file(self.source_image_path)

        m_exists.assert_called_once_with(self.source_image_path)
        m_get_osgeo_path.assert_called_once_with(self.source_image_path)
        m_gdal_open.assert_called_once_with(self.osgeo_image_path)

    @mock.patch("rok4.raster.get_osgeo_path")
    @mock.patch("rok4.raster.gdal.IdentifyDriver")
    @mock.patch("rok4.raster.gdal.Open", side_effect=None)
    @mock.patch("rok4.raster.exists", side_effect=[True, True])
    def test_unsupported_mask_format(
        self, m_exists, m_gdal_open, m_identifydriver, m_get_osgeo_path
    ):
        """Test case : Constructor called on an unsupported mask file or object."""
        m_get_osgeo_path.side_effect = [self.osgeo_image_path, self.osgeo_mask_path]
        m_identifydriver.return_value = type("", (object,), {"ShortName": "JPG"})

        with pytest.raises(Exception):
            Raster.from_file(self.source_image_path)

        m_exists.assert_has_calls([call(self.source_image_path), call(self.source_mask_path)])
        m_get_osgeo_path.assert_has_calls(
            [call(self.source_image_path), call(self.source_mask_path)]
        )
        m_identifydriver.assert_called_once_with(self.osgeo_mask_path)
        m_gdal_open.assert_called_once_with(self.osgeo_image_path)


class TestRasterFromParameters(TestCase):
    """rok4.raster.Raster.from_parameters(**kwargs) class constructor."""

    def test_image(self):
        """Parameters describing an image without mask"""
        parameters = {
            "bands": 4,
            "bbox": (-5.4, 41.3, 9.8, 51.3),
            "dimensions": (1920, 1080),
            "format": ColorFormat.UINT8,
            "path": "file:///path/to/image.tif",
        }

        raster = Raster.from_parameters(**parameters)

        assert raster.path == parameters["path"]
        assert (
            isinstance(raster.bbox, tuple)
            and len(raster.bbox) == 4
            and math.isclose(raster.bbox[0], parameters["bbox"][0], rel_tol=1e-5)
            and math.isclose(raster.bbox[1], parameters["bbox"][1], rel_tol=1e-5)
            and math.isclose(raster.bbox[2], parameters["bbox"][2], rel_tol=1e-5)
            and math.isclose(raster.bbox[3], parameters["bbox"][3], rel_tol=1e-5)
        )
        assert raster.bands == parameters["bands"]
        assert raster.format == parameters["format"]
        assert raster.dimensions == parameters["dimensions"]
        assert raster.mask is None

    def test_image_and_mask(self):
        """Parameters describing an image with mask"""
        parameters = {
            "bands": 4,
            "bbox": (-5.4, 41.3, 9.8, 51.3),
            "dimensions": (1920, 1080),
            "format": ColorFormat.UINT8,
            "mask": "file:///path/to/image.msk",
            "path": "file:///path/to/image.tif",
        }

        raster = Raster.from_parameters(**parameters)

        assert raster.path == parameters["path"]
        assert (
            isinstance(raster.bbox, tuple)
            and len(raster.bbox) == 4
            and math.isclose(raster.bbox[0], parameters["bbox"][0], rel_tol=1e-5)
            and math.isclose(raster.bbox[1], parameters["bbox"][1], rel_tol=1e-5)
            and math.isclose(raster.bbox[2], parameters["bbox"][2], rel_tol=1e-5)
            and math.isclose(raster.bbox[3], parameters["bbox"][3], rel_tol=1e-5)
        )
        assert raster.bands == parameters["bands"]
        assert raster.format == parameters["format"]
        assert raster.dimensions == parameters["dimensions"]
        assert raster.mask == parameters["mask"]


# rok4.raster.RasterSet class tests


class TestRasterSetInit(TestCase):
    """rok4.raster.RasterSet default constructor."""

    def test_default(self):
        """Default property values."""
        rasterset = RasterSet()

        assert (
            isinstance(rasterset.bbox, tuple)
            and len(rasterset.bbox) == 4
            and all(coordinate is None for coordinate in rasterset.bbox)
        )
        assert isinstance(rasterset.colors, list) and not rasterset.colors
        assert isinstance(rasterset.raster_list, list) and not rasterset.raster_list
        assert rasterset.srs is None


class TestRasterSetFromList(TestCase):
    """rok4.raster.RasterSet.from_list(path, srs) class constructor."""

    @mock.patch("rok4.raster.get_osgeo_path")
    @mock.patch("rok4.raster.Raster.from_file")
    def test_ok_at_least_3_files(self, m_from_file, m_get_osgeo_path):
        """List of 3 or more valid image files"""
        file_number = random.randint(3, 100)
        file_list = []
        for n in range(0, file_number, 1):
            file_list.append(f"s3://test_bucket/image_{n+1}.tif")
        file_list_string = "\n".join(file_list)
        m_open = mock_open(read_data=file_list_string)
        list_path = "s3://test_bucket/raster_set.list"
        list_local_path = "/tmp/raster_set.list"
        m_get_osgeo_path.return_value = list_local_path
        raster_list = []
        colors = []
        serial_in = {"raster_list": [], "colors": []}
        for n in range(0, file_number, 1):
            raster = MagicMock(Raster)
            raster.path = file_list[n]
            raster.bbox = (
                -0.75 + math.floor(n / 3),
                -1.33 + n - 3 * math.floor(n / 3),
                0.25 + math.floor(n / 3),
                -0.33 + n - 3 * math.floor(n / 3),
            )
            raster.format = random.choice([ColorFormat.BIT, ColorFormat.UINT8, ColorFormat.FLOAT32])
            if raster.format == ColorFormat.BIT:
                raster.bands = 1
            else:
                raster.bands = random.randint(1, 4)
            if random.randint(0, 1) == 1:
                raster.mask = raster.path.replace(".tif", ".msk")
            else:
                raster.mask = None
            color_dict = {"bands": raster.bands, "format": raster.format}
            if color_dict not in colors:
                colors.append(color_dict)
                serial_in["colors"].append({"bands": raster.bands, "format": raster.format.name})
            raster.dimensions = (5000, 5000)
            raster_list.append(raster)
            raster_serial = {
                "path": raster.path,
                "bands": raster.bands,
                "format": raster.format.name,
                "bbox": list(raster.bbox),
                "dimensions": list(raster.dimensions),
            }
            if raster.mask:
                raster_serial["mask"] = raster.mask
            serial_in["raster_list"].append(raster_serial)
        m_from_file.side_effect = raster_list
        srs = "EPSG:4326"
        serial_in["srs"] = srs
        bbox = (-0.75, -1.33, 0.25 + math.floor((file_number - 1) / 3), 1.67)
        serial_in["bbox"] = list(bbox)

        with mock.patch("rok4.raster.open", m_open):
            rasterset = RasterSet.from_list(list_path, srs)

        serial_out = rasterset.serializable
        assert rasterset.srs == srs
        m_get_osgeo_path.assert_called_once_with(list_path)
        m_open.assert_called_once_with(file=list_local_path)
        assert rasterset.raster_list == raster_list
        assert isinstance(serial_out["bbox"], list)
        for i in range(0, 4, 1):
            assert math.isclose(rasterset.bbox[i], bbox[i], rel_tol=1e-5)
            assert math.isclose(serial_out["bbox"][i], serial_in["bbox"][i], rel_tol=1e-5)
        assert len(rasterset.colors) > 0
        assert rasterset.colors == colors
        for key in serial_in.keys():
            if key != "bbox":
                assert serial_out[key] == serial_in[key]
        assert isinstance(serial_out["bbox"], list)


class TestRasterSetFromDescriptor(TestCase):
    """rok4.raster.RasterSet.from_descriptor(path) class constructor."""

    @mock.patch("rok4.raster.get_osgeo_path")
    @mock.patch("rok4.raster.Raster.from_parameters")
    def test_simple_ok(self, m_from_parameters, m_get_osgeo_path):
        serial_in = {
            "bbox": [550000.000, 6210000.000, 570000.000, 6230000.000],
            "colors": [{"bands": 3, "format": "UINT8"}],
            "raster_list": [
                {
                    "bands": 3,
                    "bbox": [550000.000, 6210000.000, 560000.000, 6220000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "file:///path/to/images/550000_6220000.msk",
                    "path": "file:///path/to/images/550000_6220000.tif",
                },
                {
                    "bands": 3,
                    "bbox": [560000.000, 6210000.000, 570000.000, 6220000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "file:///path/to/images/560000_6220000.msk",
                    "path": "file:///path/to/images/560000_6220000.tif",
                },
                {
                    "bands": 3,
                    "bbox": [550000.000, 6220000.000, 560000.000, 6230000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "file:///path/to/images/550000_6230000.msk",
                    "path": "file:///path/to/images/550000_6230000.tif",
                },
            ],
            "srs": "IGNF:LAMB93",
        }
        desc_path = "file:///path/to/descriptor.json"
        local_path = "/path/to/descriptor.json"
        desc_content = json.dumps(serial_in)
        raster_list = []
        raster_args_list = []
        for raster_dict in serial_in["raster_list"]:
            raster_properties = copy.deepcopy(raster_dict)
            raster_properties["format"] = ColorFormat[raster_dict["format"]]
            raster_properties["bbox"] = tuple(raster_dict["bbox"])
            raster_properties["dimensions"] = tuple(raster_dict["dimensions"])

            raster = MagicMock(Raster, **raster_properties)
            raster_list.append(raster)
            raster_args_list.append(raster_properties)
        m_from_parameters.side_effect = raster_list
        m_get_osgeo_path.return_value = local_path
        m_open = mock_open(read_data=desc_content)

        with mock.patch("rok4.raster.open", m_open):
            rasterset = RasterSet.from_descriptor(desc_path)

        m_get_osgeo_path.assert_called_once_with(desc_path)
        m_open.assert_called_once_with(file=local_path)
        assert rasterset.srs == serial_in["srs"]
        m_from_parameters.assert_called()
        assert m_from_parameters.call_count == 3
        for i in range(0, len(raster_args_list), 1):
            assert m_from_parameters.call_args_list[i] == call(**raster_args_list[i])
        assert rasterset.raster_list == raster_list
        assert isinstance(rasterset.bbox, tuple) and len(rasterset.bbox) == 4
        assert isinstance(rasterset.colors, list) and rasterset.colors
        for i in range(0, len(serial_in["colors"]), 1):
            expected_color = copy.deepcopy(serial_in["colors"][i])
            expected_color["format"] = ColorFormat[serial_in["colors"][i]["format"]]
            assert rasterset.colors[i] == expected_color
        serial_out = rasterset.serializable
        assert isinstance(serial_out["bbox"], list) and len(serial_out["bbox"]) == 4
        for i in range(0, 4, 1):
            assert math.isclose(rasterset.bbox[i], serial_in["bbox"][i], rel_tol=1e-5)
            assert math.isclose(serial_out["bbox"][i], serial_in["bbox"][i], rel_tol=1e-5)
        for key in serial_in.keys():
            if key != "bbox":
                assert serial_out[key] == serial_in[key]


class TestRasterSetWriteDescriptor(TestCase):
    """rok4.raster.RasterSet.write_descriptor(path) class method."""

    @mock.patch("rok4.raster.put_data_str")
    def test_ok_with_output_path(self, m_put_data_str):
        serial_in = {
            "bbox": [550000.000, 6210000.000, 570000.000, 6230000.000],
            "colors": [{"bands": 3, "format": "UINT8"}],
            "raster_list": [
                {
                    "bands": 3,
                    "bbox": [550000.000, 6210000.000, 560000.000, 6220000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "s3://rok4bucket/images/550000_6220000.msk",
                    "path": "s3://rok4bucket/images/550000_6220000.tif",
                },
                {
                    "bands": 3,
                    "bbox": [560000.000, 6210000.000, 570000.000, 6220000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "s3://rok4bucket/images/560000_6220000.msk",
                    "path": "s3://rok4bucket/images/560000_6220000.tif",
                },
                {
                    "bands": 3,
                    "bbox": [550000.000, 6220000.000, 560000.000, 6230000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "s3://rok4bucket/images/550000_6230000.msk",
                    "path": "s3://rok4bucket/images/550000_6230000.tif",
                },
            ],
            "srs": "IGNF:LAMB93",
        }
        content = json.dumps(serial_in, sort_keys=True)
        path = "s3://rok4bucket/dst_descriptor.json"
        rasterset = RasterSet()
        rasterset.bbox = tuple(serial_in["bbox"])
        rasterset.srs = serial_in["srs"]
        rasterset.colors = []
        for color_dict in serial_in["colors"]:
            rasterset.colors.append(
                {"bands": color_dict["bands"], "format": ColorFormat[color_dict["format"]]}
            )
        rasterset.raster_list = []
        for raster_dict in serial_in["raster_list"]:
            raster_args = copy.deepcopy(raster_dict)
            raster_args["format"] = ColorFormat[raster_dict["format"]]
            raster_args["bbox"] = tuple(raster_dict["bbox"])
            raster_args["dimensions"] = tuple(raster_dict["dimensions"])
            rasterset.raster_list.append(MagicMock(Raster, **raster_args))

        try:
            rasterset.write_descriptor(path)
        except Exception as exc:
            assert False, f"Writing RasterSet's descriptor raises an exception: {exc}"

        m_put_data_str.assert_called_once_with(content, path)

    @mock.patch("rok4.raster.print")
    def test_ok_no_output_path(self, m_print):
        serial_in = {
            "bbox": [550000.000, 6210000.000, 570000.000, 6230000.000],
            "colors": [{"bands": 3, "format": "UINT8"}],
            "raster_list": [
                {
                    "bands": 3,
                    "bbox": [550000.000, 6210000.000, 560000.000, 6220000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "s3://rok4bucket/images/550000_6220000.msk",
                    "path": "s3://rok4bucket/images/550000_6220000.tif",
                },
                {
                    "bands": 3,
                    "bbox": [560000.000, 6210000.000, 570000.000, 6220000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "s3://rok4bucket/images/560000_6220000.msk",
                    "path": "s3://rok4bucket/images/560000_6220000.tif",
                },
                {
                    "bands": 3,
                    "bbox": [550000.000, 6220000.000, 560000.000, 6230000.000],
                    "dimensions": [5000, 5000],
                    "format": "UINT8",
                    "mask": "s3://rok4bucket/images/550000_6230000.msk",
                    "path": "s3://rok4bucket/images/550000_6230000.tif",
                },
            ],
            "srs": "IGNF:LAMB93",
        }
        content = json.dumps(serial_in, sort_keys=True)
        rasterset = RasterSet()
        rasterset.bbox = tuple(serial_in["bbox"])
        rasterset.srs = serial_in["srs"]
        rasterset.colors = []
        for color_dict in serial_in["colors"]:
            rasterset.colors.append(
                {"bands": color_dict["bands"], "format": ColorFormat[color_dict["format"]]}
            )
        rasterset.raster_list = []
        for raster_dict in serial_in["raster_list"]:
            raster_args = copy.deepcopy(raster_dict)
            raster_args["format"] = ColorFormat[raster_dict["format"]]
            raster_args["bbox"] = tuple(raster_dict["bbox"])
            raster_args["dimensions"] = tuple(raster_dict["dimensions"])
            rasterset.raster_list.append(MagicMock(Raster, **raster_args))

        try:
            rasterset.write_descriptor()
        except Exception as exc:
            assert False, f"Writing RasterSet's descriptor raises an exception: {exc}"

        m_print.assert_called_once_with(content)



"""Provide classes to use pyramid's data.

The module contains the following classes:

- `Pyramid` - Data container
- `Level` - Level of a pyramid
"""

# -- IMPORTS --

# standard library
import io
import json
import os
import re
import tempfile
import zlib
from json.decoder import JSONDecodeError
from typing import Dict, Iterator, List, Tuple

# 3rd party
import mapbox_vector_tile
import numpy
from PIL import Image

# package
from rok4.enums import PyramidType, SlabType, StorageType
from rok4.exceptions import FormatError, MissingAttributeError
from rok4.storage import (
    copy,
    get_data_binary,
    get_data_str,
    get_infos_from_path,
    get_path_from_infos,
    put_data_str,
    remove,
    size_path,
)
from rok4.tile_matrix_set import TileMatrix, TileMatrixSet
from rok4.utils import reproject_point, srs_to_spatialreference

# -- GLOBALS --
ROK4_IMAGE_HEADER_SIZE = 2048
"""Slab's header size, 2048 bytes"""


def b36_number_encode(number: int) -> str:
    """Convert base-10 number to base-36

    Used alphabet is '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'

    Args:
        number (int): base-10 number

    Returns:
        str: base-36 number
    """

    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    base36 = ""

    if 0 <= number < len(alphabet):
        return alphabet[number]

    while number != 0:
        number, i = divmod(number, len(alphabet))
        base36 = alphabet[i] + base36

    return base36


def b36_number_decode(number: str) -> int:
    """Convert base-36 number to base-10

    Args:
        number (str): base-36 number

    Returns:
        int: base-10 number
    """
    return int(number, 36)


def b36_path_decode(path: str) -> Tuple[int, int]:
    """Get slab's column and row from a base-36 based path

    Args:
        path (str): slab's path

    Returns:
        Tuple[int, int]: slab's column and row
    """

    path = path.replace("/", "")
    path = re.sub(r"(\.TIFF?)", "", path.upper())

    b36_column = ""
    b36_row = ""

    while len(path) > 0:
        b36_column += path[0]
        b36_row += path[1]
        path = path[2:]

    return b36_number_decode(b36_column), b36_number_decode(b36_row)


def b36_path_encode(column: int, row: int, slashs: int) -> str:
    """Convert slab indices to base-36 based path, with .tif extension

    Args:
        column (int): slab's column
        row (int): slab's row
        slashs (int): slashs' number (to split path)

    Returns:
        str: base-36 based path
    """

    b36_column = b36_number_encode(column)
    b36_row = b36_number_encode(row)

    max_len = max(slashs + 1, len(b36_column), len(b36_row))

    b36_column = b36_column.rjust(max_len, "0")
    b36_row = b36_row.rjust(max_len, "0")

    b36_path = ""

    while len(b36_column) > 0:
        b36_path = b36_row[-1] + b36_path
        b36_path = b36_column[-1] + b36_path

        b36_column = b36_column[:-1]
        b36_row = b36_row[:-1]

        if slashs > 0:
            b36_path = "/" + b36_path
            slashs -= 1

    return f"{b36_path}.tif"


class Level:
    """A pyramid's level, raster or vector

    Attributes:
        __id (str): level's identifier. have to exist in the pyramid's used TMS
        __tile_limits (Dict[str, int]): minimum and maximum tiles' columns and rows of pyramid's content
        __slab_size (Tuple[int, int]): number of tile in a slab, widthwise and heightwise
        __tables (List[Dict]): for a VECTOR pyramid, description of vector content, tables and attributes
    """

    @classmethod
    def from_descriptor(cls, data: Dict, pyramid: "Pyramid") -> "Level":
        """Create a pyramid's level from the pyramid's descriptor levels element

        Args:
            data (Dict): level's information from the pyramid's descriptor
            pyramid (Pyramid): pyramid containing the level to create

        Raises:
            Exception: different storage or masks presence between the level and the pyramid
            MissingAttributeError: Attribute is missing in the content

        Returns:
            Pyramid: a Level instance
        """
        level = cls()

        level.__pyramid = pyramid

        # Attributs communs
        try:
            level.__id = data["id"]
            level.__tile_limits = data["tile_limits"]
            level.__slab_size = (
                data["tiles_per_width"],
                data["tiles_per_height"],
            )

            # Informations sur le stockage : on les valide et stocke dans la pyramide
            if pyramid.storage_type.name != data["storage"]["type"]:
                raise Exception(
                    f"Pyramid {pyramid.descriptor} owns levels using different storage types ({ data['storage']['type'] }) than its one ({pyramid.storage_type.name})"
                )

            if pyramid.storage_type == StorageType.FILE:
                pyramid.storage_depth = data["storage"]["path_depth"]

            if "mask_directory" in data["storage"] or "mask_prefix" in data["storage"]:
                if not pyramid.own_masks:
                    raise Exception(
                        f"Pyramid {pyramid.__descriptor} does not define a mask format but level {level.__id} define mask storage informations"
                    )
            else:
                if pyramid.own_masks:
                    raise Exception(
                        f"Pyramid {pyramid.__descriptor} define a mask format but level {level.__id} does not define mask storage informations"
                    )

        except KeyError as e:
            raise MissingAttributeError(pyramid.descriptor, f"levels[].{e}")

        # Attributs dans le cas d'un niveau vecteur
        if level.__pyramid.type == PyramidType.VECTOR:
            try:
                level.__tables = data["tables"]

            except KeyError as e:
                raise MissingAttributeError(pyramid.descriptor, f"levels[].{e}")

        return level

    @classmethod
    def from_other(cls, other: "Level", pyramid: "Pyramid") -> "Level":
        """Create a pyramid's level from another one

        Args:
            other (Level): level to clone
            pyramid (Pyramid): new pyramid containing the new level

        Raises:
            Exception: different storage or masks presence between the level and the pyramid
            MissingAttributeError: Attribute is missing in the content

        Returns:
            Pyramid: a Level instance
        """

        level = cls()

        # Attributs communs
        level.__id = other.__id
        level.__pyramid = pyramid
        level.__tile_limits = other.__tile_limits
        level.__slab_size = other.__slab_size

        # Attributs dans le cas d'un niveau vecteur
        if level.__pyramid.type == PyramidType.VECTOR:
            level.__tables = other.__tables

        return level

    def __str__(self) -> str:
        return f"{self.__pyramid.type.name} pyramid's level '{self.__id}' ({self.__pyramid.storage_type.name} storage)"

    @property
    def serializable(self) -> Dict:
        """Get the dict version of the pyramid object, pyramid's descriptor compliant

        Returns:
            Dict: pyramid's descriptor structured object description
        """
        serialization = {
            "id": self.__id,
            "tiles_per_width": self.__slab_size[0],
            "tiles_per_height": self.__slab_size[1],
            "tile_limits": self.__tile_limits,
        }

        if self.__pyramid.type == PyramidType.VECTOR:
            serialization["tables"] = self.__tables

        if self.__pyramid.storage_type == StorageType.FILE:
            serialization["storage"] = {
                "type": "FILE",
                "image_directory": f"{self.__pyramid.name}/DATA/{self.__id}",
                "path_depth": self.__pyramid.storage_depth,
            }
            if self.__pyramid.own_masks:
                serialization["storage"][
                    "mask_directory"
                ] = f"{self.__pyramid.name}/MASK/{self.__id}"

        elif self.__pyramid.storage_type == StorageType.CEPH:
            serialization["storage"] = {
                "type": "CEPH",
                "image_prefix": f"{self.__pyramid.name}/DATA_{self.__id}",
                "pool_name": self.__pyramid.storage_root,
            }
            if self.__pyramid.own_masks:
                serialization["storage"]["mask_prefix"] = f"{self.__pyramid.name}/MASK_{self.__id}"

        elif self.__pyramid.storage_type == StorageType.S3:
            serialization["storage"] = {
                "type": "S3",
                "image_prefix": f"{self.__pyramid.name}/DATA_{self.__id}",
                "bucket_name": self.__pyramid.storage_root,
            }
            if self.__pyramid.own_masks:
                serialization["storage"]["mask_prefix"] = f"{self.__pyramid.name}/MASK_{self.__id}"

        return serialization

    @property
    def id(self) -> str:
        return self.__id

    @property
    def bbox(self) -> Tuple[float, float, float, float]:
        """Return level extent, based on tile limits

        Returns:
            Tuple[float, float, float, float]: level terrain extent (xmin, ymin, xmax, ymax)
        """

        min_bbox = self.__pyramid.tms.get_level(self.__id).tile_to_bbox(
            self.__tile_limits["min_col"], self.__tile_limits["max_row"]
        )
        max_bbox = self.__pyramid.tms.get_level(self.__id).tile_to_bbox(
            self.__tile_limits["max_col"], self.__tile_limits["min_row"]
        )

        return (min_bbox[0], min_bbox[1], max_bbox[2], max_bbox[3])

    @property
    def resolution(self) -> str:
        return self.__pyramid.tms.get_level(self.__id).resolution

    @property
    def tile_matrix(self) -> TileMatrix:
        return self.__pyramid.tms.get_level(self.__id)

    @property
    def slab_width(self) -> int:
        return self.__slab_size[0]

    @property
    def slab_height(self) -> int:
        return self.__slab_size[1]

    def is_in_limits(self, column: int, row: int) -> bool:
        """Is the tile indices in limits ?

        Args:
            column (int): tile's column
            row (int): tile's row

        Returns:
            bool: True if tiles' limits contain the provided tile's indices
        """
        return (
            self.__tile_limits["min_row"] <= row
            and self.__tile_limits["max_row"] >= row
            and self.__tile_limits["min_col"] <= column
            and self.__tile_limits["max_col"] >= column
        )

    def set_limits_from_bbox(self, bbox: Tuple[float, float, float, float]) -> None:
        """Set tile limits, based on provided bounding box

        Args:
            bbox (Tuple[float, float, float, float]): terrain extent (xmin, ymin, xmax, ymax), in TMS coordinates system

        """

        col_min, row_min, col_max, row_max = self.__pyramid.tms.get_level(self.__id).bbox_to_tiles(
            bbox
        )
        self.__tile_limits = {
            "min_row": row_min,
            "max_col": col_max,
            "max_row": row_max,
            "min_col": col_min,
        }


class Pyramid:

    """A data pyramid, raster or vector

    Attributes:
        __name (str): pyramid's name
        __descriptor (str): pyramid's descriptor path
        __list (str): pyramid's list path
        __tms (rok4.tile_matrix_set.TileMatrixSet): Used grid
        __levels (Dict[str, Level]): Pyramid's levels
        __format (str): Data format
        __storage (Dict[str, Union[rok4.storage.StorageType,str,int]]): Pyramid's storage informations (type, root and depth if FILE storage)
        __raster_specifications (Dict): If raster pyramid, raster specifications
        __content (Dict): Loading status (loaded) and list content (cache).

            Example (S3 storage):

                {
                    'cache': {
                        (<SlabType.DATA: 'DATA'>, '18', 5424, 7526): {
                            'link': False,
                            'md5': None,
                            'root': 'pyramids@localhost:9000/LIMADM',
                            'slab': 'DATA_18_5424_7526'
                        }
                    },
                    'loaded': True
                }
    """

    @classmethod
    def from_descriptor(cls, descriptor: str) -> "Pyramid":
        """Create a pyramid from its descriptor

        Args:
            descriptor (str): pyramid's descriptor path

        Raises:
            FormatError: Provided path or the TMS is not a well formed JSON
            Exception: Level issue : no one in the pyramid or the used TMS, or level ID not defined in the TMS
            MissingAttributeError: Attribute is missing in the content
            StorageError: Storage read issue (pyramid descriptor or TMS)
            MissingEnvironmentError: Missing object storage informations or TMS root directory

        Examples:

            S3 stored descriptor

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("s3://bucket_name/path/to/descriptor.json")
                except Exception as e:
                    print("Cannot load the pyramid from its descriptor")

        Returns:
            Pyramid: a Pyramid instance
        """
        try:
            data = json.loads(get_data_str(descriptor))

        except JSONDecodeError as e:
            raise FormatError("JSON", descriptor, e)

        pyramid = cls()

        pyramid.__storage["type"], path, pyramid.__storage["root"], base_name = get_infos_from_path(
            descriptor
        )
        pyramid.__name = base_name[:-5]  # on supprime l'extension.json
        pyramid.__descriptor = descriptor
        pyramid.__list = get_path_from_infos(
            pyramid.__storage["type"], pyramid.__storage["root"], f"{pyramid.__name}.list"
        )

        try:
            # Attributs communs
            pyramid.__tms = TileMatrixSet(data["tile_matrix_set"])
            pyramid.__format = data["format"]

            # Attributs d'une pyramide raster
            if pyramid.type == PyramidType.RASTER:
                pyramid.__raster_specifications = data["raster_specifications"]

                if "mask_format" in data:
                    pyramid.__masks = True
                else:
                    pyramid.__masks = False

            # Niveaux
            for level in data["levels"]:
                lev = Level.from_descriptor(level, pyramid)
                pyramid.__levels[lev.id] = lev

                if pyramid.__tms.get_level(lev.id) is None:
                    raise Exception(
                        f"Pyramid {descriptor} owns a level with the ID '{lev.id}', not defined in the TMS '{pyramid.tms.name}'"
                    )

        except KeyError as e:
            raise MissingAttributeError(descriptor, e)

        if len(pyramid.__levels.keys()) == 0:
            raise Exception(f"Pyramid '{descriptor}' has no level")

        return pyramid

    @classmethod
    def from_other(cls, other: "Pyramid", name: str, storage: Dict) -> "Pyramid":
        """Create a pyramid from another one

        Args:
            other (Pyramid): pyramid to clone
            name (str): new pyramid's name
            storage (Dict[str, Union[str, int]]): new pyramid's storage informations

        Raises:
            FormatError: Provided path or the TMS is not a well formed JSON
            Exception: Level issue : no one in the pyramid or the used TMS, or level ID not defined in the TMS
            MissingAttributeError: Attribute is missing in the content

        Returns:
            Pyramid: a Pyramid instance
        """
        try:
            # On convertit le type de stockage selon l'énumération
            storage["type"] = StorageType[storage["type"]]

            if storage["type"] == StorageType.FILE and name.find("/") != -1:
                raise Exception(f"A FILE stored pyramid's name cannot contain '/' : '{name}'")

            if storage["type"] == StorageType.FILE and "depth" not in storage:
                storage["depth"] = 2

            pyramid = cls()

            # Attributs communs
            pyramid.__name = name
            pyramid.__storage = storage
            pyramid.__masks = other.__masks

            pyramid.__descriptor = get_path_from_infos(
                pyramid.__storage["type"], pyramid.__storage["root"], f"{pyramid.__name}.json"
            )
            pyramid.__list = get_path_from_infos(
                pyramid.__storage["type"], pyramid.__storage["root"], f"{pyramid.__name}.list"
            )
            pyramid.__tms = other.__tms
            pyramid.__format = other.__format

            # Attributs d'une pyramide raster
            if pyramid.type == PyramidType.RASTER:
                if other.own_masks:
                    pyramid.__masks = True
                else:
                    pyramid.__masks = False
                pyramid.__raster_specifications = other.__raster_specifications

            # Niveaux
            for level in other.__levels.values():
                lev = Level.from_other(level, pyramid)
                pyramid.__levels[lev.id] = lev

        except KeyError as e:
            raise MissingAttributeError(pyramid.descriptor, e)

        return pyramid

    def __init__(self) -> None:
        self.__storage = {}
        self.__levels = {}
        self.__masks = None

        self.__content = {"loaded": False, "cache": {}}

    def __str__(self) -> str:
        return f"{self.type.name} pyramid '{self.__name}' ({self.__storage['type'].name} storage)"

    @property
    def serializable(self) -> Dict:
        """Get the dict version of the pyramid object, descriptor compliant

        Returns:
            Dict: descriptor structured object description
        """

        serialization = {"tile_matrix_set": self.__tms.name, "format": self.__format}

        serialization["levels"] = []
        sorted_levels = sorted(
            self.__levels.values(), key=lambda level: level.resolution, reverse=True
        )

        for level in sorted_levels:
            serialization["levels"].append(level.serializable)

        if self.type == PyramidType.RASTER:
            serialization["raster_specifications"] = self.__raster_specifications

        if self.__masks:
            serialization["mask_format"] = "TIFF_ZIP_UINT8"

        return serialization

    @property
    def list(self) -> str:
        return self.__list

    @property
    def descriptor(self) -> str:
        return self.__descriptor

    @property
    def name(self) -> str:
        return self.__name

    @property
    def tms(self) -> TileMatrixSet:
        return self.__tms

    @property
    def raster_specifications(self) -> Dict:
        """Get raster specifications for a RASTER pyramid

        Example:
            {
                "channels": 3,
                "nodata": "255,0,0",
                "photometric": "rgb",
                "interpolation": "bicubic"
            }

        Returns:
            Dict: Raster specifications, None if VECTOR pyramid
        """
        return self.__raster_specifications

    @property
    def storage_type(self) -> StorageType:
        """Get the storage type

        Returns:
            StorageType: FILE, S3 or CEPH
        """
        return self.__storage["type"]

    @property
    def storage_root(self) -> str:
        """Get the pyramid's storage root.

        If storage is S3, the used cluster is removed.

        Returns:
            str: Pyramid's storage root
        """

        return self.__storage["root"].split("@", 1)[
            0
        ]  # Suppression de l'éventuel hôte de spécification du cluster S3

    @property
    def storage_depth(self) -> int:
        return self.__storage.get("depth", None)

    @property
    def storage_s3_cluster(self) -> str:
        """Get the pyramid's storage S3 cluster (host name)

        Returns:
            str: the host if known, None if the default one have to be used or if storage is not S3
        """
        if self.__storage["type"] == StorageType.S3:
            try:
                return self.__storage["root"].split("@")[1]
            except IndexError:
                return None
        else:
            return None

    @storage_depth.setter
    def storage_depth(self, d: int) -> None:
        """Set the tree depth for a FILE storage

        Args:
            d (int): file storage depth

        Raises:
            Exception: the depth is not equal to the already known depth
        """
        if "depth" in self.__storage and self.__storage["depth"] != d:
            raise Exception(f"Pyramid {self.__descriptor} owns levels with different path depths")
        self.__storage["depth"] = d

    @property
    def own_masks(self) -> bool:
        return self.__masks

    @property
    def format(self) -> str:
        return self.__format

    @property
    def tile_extension(self) -> str:
        if self.__format in [
            "TIFF_RAW_UINT8",
            "TIFF_LZW_UINT8",
            "TIFF_ZIP_UINT8",
            "TIFF_PKB_UINT8",
            "TIFF_RAW_FLOAT32",
            "TIFF_LZW_FLOAT32",
            "TIFF_ZIP_FLOAT32",
            "TIFF_PKB_FLOAT32",
        ]:
            return "tif"
        elif self.__format in ["TIFF_JPG_UINT8", "TIFF_JPG90_UINT8"]:
            return "jpg"
        elif self.__format == "TIFF_PNG_UINT8":
            return "png"
        elif self.__format == "TIFF_PBF_MVT":
            return "pbf"
        else:
            raise Exception(
                f"Unknown pyramid's format ({self.__format}), cannot return the tile extension"
            )

    @property
    def bottom_level(self) -> "Level":
        """Get the best resolution level in the pyramid

        Returns:
            Level: the bottom level
        """
        return sorted(self.__levels.values(), key=lambda level: level.resolution)[0]

    @property
    def top_level(self) -> "Level":
        """Get the low resolution level in the pyramid

        Returns:
            Level: the top level
        """
        return sorted(self.__levels.values(), key=lambda level: level.resolution)[-1]

    @property
    def type(self) -> PyramidType:
        """Get the pyramid's type (RASTER or VECTOR) from its format

        Returns:
            PyramidType: RASTER or VECTOR
        """
        if self.__format == "TIFF_PBF_MVT":
            return PyramidType.VECTOR
        else:
            return PyramidType.RASTER

    def load_list(self) -> None:
        """Load list content and cache it

        If list is already loaded, nothing done
        """
        if self.__content["loaded"]:
            return

        for slab, infos in self.list_generator():
            self.__content["cache"][slab] = infos

        self.__content["loaded"] = True

    def list_generator(self) -> Iterator[Tuple[Tuple[SlabType, str, int, int], Dict]]:
        """Get list content

        List is copied as temporary file, roots are read and informations about each slab is returned. If list is already loaded, we yield the cached content

        Examples:

            S3 stored descriptor

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("s3://bucket_name/path/to/descriptor.json")

                    for (slab_type, level, column, row), infos in pyramid.list_generator():
                        print(infos)

                except Exception as e:
                    print("Cannot load the pyramid from its descriptor and read the list")

        Yields:
            Iterator[Tuple[Tuple[SlabType,str,int,int], Dict]]: Slab indices and storage informations

            Value example:

                (
                    (<SlabType.DATA: 'DATA'>, '18', 5424, 7526),
                    {
                        'link': False,
                        'md5': None,
                        'root': 'pyramids@localhost:9000/LIMADM',
                        'slab': 'DATA_18_5424_7526'
                    }
                )

        Raises:
            StorageError: Unhandled pyramid storage to copy list
            MissingEnvironmentError: Missing object storage informations
        """
        if self.__content["loaded"]:
            for slab, infos in self.__content["cache"].items():
                yield slab, infos
        else:
            # Copie de la liste dans un fichier temporaire (cette liste peut être un objet)
            list_obj = tempfile.NamedTemporaryFile(mode="r", delete=False)
            list_file = list_obj.name
            copy(self.__list, f"file://{list_file}")
            list_obj.close()

            roots = {}
            s3_cluster = self.storage_s3_cluster

            with open(list_file) as listin:
                # Lecture des racines
                for line in listin:
                    line = line.rstrip()

                    if line == "#":
                        break

                    root_id, root_path = line.split("=", 1)

                    if s3_cluster is None:
                        roots[root_id] = root_path
                    else:
                        # On a un nom de cluster S3, on l'ajoute au nom du bucket dans les racines
                        root_bucket, root_path = root_path.split("/", 1)
                        roots[root_id] = f"{root_bucket}@{s3_cluster}/{root_path}"

                # Lecture des dalles
                for line in listin:
                    line = line.rstrip()

                    parts = line.split(" ", 1)
                    slab_path = parts[0]
                    slab_md5 = None
                    if len(parts) == 2:
                        slab_md5 = parts[1]

                    root_id, slab_path = slab_path.split("/", 1)

                    slab_type, level, column, row = self.get_infos_from_slab_path(slab_path)
                    infos = {
                        "root": roots[root_id],
                        "link": root_id != "0",
                        "slab": slab_path,
                        "md5": slab_md5,
                    }

                    yield ((slab_type, level, column, row), infos)

            remove(f"file://{list_file}")

    def get_level(self, level_id: str) -> "Level":
        """Get one level according to its identifier

        Args:
            level_id: Level identifier

        Returns:
            The corresponding pyramid's level, None if not present
        """

        return self.__levels.get(level_id, None)

    def get_levels(self, bottom_id: str = None, top_id: str = None) -> List[Level]:
        """Get sorted levels in the provided range from bottom to top

        Args:
            bottom_id (str, optionnal): specific bottom level id. Defaults to None.
            top_id (str, optionnal): specific top level id. Defaults to None.

        Raises:
            Exception: Provided levels are not consistent (bottom > top or not in the pyramid)

        Examples:

            All levels

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("s3://bucket_name/path/to/descriptor.json")
                    levels = pyramid.get_levels()

                except Exception as e:
                    print("Cannot load the pyramid from its descriptor and get levels")

            From pyramid's bottom to provided top (level 5)

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("s3://bucket_name/path/to/descriptor.json")
                    levels = pyramid.get_levels(None, "5")

                except Exception as e:
                    print("Cannot load the pyramid from its descriptor and get levels")

        Returns:
            List[Level]: asked sorted levels
        """

        sorted_levels = sorted(self.__levels.values(), key=lambda level: level.resolution)

        levels = []

        begin = False
        if bottom_id is None:
            # Pas de niveau du bas fourni, on commence tout en bas
            begin = True
        else:
            if self.get_level(bottom_id) is None:
                raise Exception(
                    f"Pyramid {self.name} does not contain the provided bottom level {bottom_id}"
                )

        if top_id is not None and self.get_level(top_id) is None:
            raise Exception(f"Pyramid {self.name} does not contain the provided top level {top_id}")

        end = False

        for level in sorted_levels:
            if not begin and level.id == bottom_id:
                begin = True

            if begin:
                levels.append(level)
                if top_id is not None and level.id == top_id:
                    end = True
                    break
                else:
                    continue

        if top_id is None:
            # Pas de niveau du haut fourni, on a été jusqu'en haut et c'est normal
            end = True

        if not begin or not end:
            raise Exception(
                f"Provided levels ids are not consistent to extract levels from the pyramid {self.name}"
            )

        return levels

    def write_descriptor(self) -> None:
        """Write the pyramid's descriptor to the final location (in the pyramid's storage root)"""

        content = json.dumps(self.serializable)
        put_data_str(content, self.__descriptor)

    def get_infos_from_slab_path(self, path: str) -> Tuple[SlabType, str, int, int]:
        """Get the slab's indices from its storage path

        Args:
            path (str): Slab's storage path

        Examples:

            FILE stored pyramid

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("/path/to/descriptor.json")
                    slab_type, level, column, row = self.get_infos_from_slab_path("DATA/12/00/4A/F7.tif")
                    # (SlabType.DATA, "12", 159, 367)
                except Exception as e:
                    print("Cannot load the pyramid from its descriptor and convert a slab path")

            S3 stored pyramid

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("s3://bucket_name/path/to/pyramid.json")
                    slab_type, level, column, row = self.get_infos_from_slab_path("s3://bucket_name/path/to/pyramid/MASK_15_9164_5846")
                    # (SlabType.MASK, "15", 9164, 5846)
                except Exception as e:
                    print("Cannot load the pyramid from its descriptor and convert a slab path")

        Returns:
            Tuple[SlabType, str, int, int]: Slab's type (DATA or MASK), level identifier, slab's column and slab's row
        """
        if self.__storage["type"] == StorageType.FILE:
            parts = path.split("/")

            # Le partie du chemin qui contient la colonne et ligne de la dalle est à la fin, en fonction de la profondeur choisie
            # depth = 2 -> on doit utiliser les 3 dernières parties pour la conversion
            column, row = b36_path_decode("/".join(parts[-(self.__storage["depth"] + 1) :]))
            level = parts[-(self.__storage["depth"] + 2)]
            raw_slab_type = parts[-(self.__storage["depth"] + 3)]

            # Pour être retro compatible avec l'ancien nommage
            if raw_slab_type == "IMAGE":
                raw_slab_type = "DATA"

            slab_type = SlabType[raw_slab_type]

            return slab_type, level, column, row
        else:
            parts = re.split(r"[/_]", path)
            column = parts[-2]
            row = parts[-1]
            level = parts[-3]
            raw_slab_type = parts[-4]

            # Pour être retro compatible avec l'ancien nommage
            if raw_slab_type == "IMG":
                raw_slab_type = "DATA"
            elif raw_slab_type == "MSK":
                raw_slab_type = "MASK"

            slab_type = SlabType[raw_slab_type]

            return slab_type, level, int(column), int(row)

    def get_slab_path_from_infos(
        self, slab_type: SlabType, level: str, column: int, row: int, full: bool = True
    ) -> str:
        """Get slab's storage path from the indices

        Args:
            slab_type (SlabType): DATA or MASK
            level (str): Level identifier
            column (int): Slab's column
            row (int): Slab's row
            full (bool, optional): Full path or just relative path from pyramid storage root. Defaults to True.

        Returns:
            str: Absolute or relative slab's storage path
        """
        if self.__storage["type"] == StorageType.FILE:
            slab_path = os.path.join(
                slab_type.value, level, b36_path_encode(column, row, self.__storage["depth"])
            )
        else:
            slab_path = f"{slab_type.value}_{level}_{column}_{row}"

        if full:
            return get_path_from_infos(
                self.__storage["type"], self.__storage["root"], self.__name, slab_path
            )
        else:
            return slab_path

    def get_tile_data_binary(self, level: str, column: int, row: int) -> str:
        """Get a pyramid's tile as binary string

        To get a tile, 3 steps :
            * calculate slab path from tile index
            * read slab index to get offsets and sizes of slab's tiles
            * read the tile into the slab

        Args:
            level (str): Tile's level
            column (int): Tile's column
            row (int): Tile's row

        Limitations:
            Pyramids with one-tile slab are not handled

        Examples:

            FILE stored raster pyramid, to extract a tile containing a point and save it as independent image

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("/data/pyramids/SCAN1000.json")
                    level, col, row, pcol, prow = pyramid.get_tile_indices(992904.46, 6733643.15, "9", srs = "IGNF:LAMB93")
                    data = pyramid.get_tile_data_binary(level, col, row)

                    if data is None:
                        print("No data")
                    else:
                        tile_name = f"tile_{level}_{col}_{row}.{pyramid.tile_extension}"
                        with open(tile_name, "wb") as image:
                            image.write(data)
                        print (f"Tile written in {tile_name}")

                except Exception as e:
                    print("Cannot save a pyramid's tile : {e}")

        Raises:
            Exception: Level not found in the pyramid
            NotImplementedError: Pyramid owns one-tile slabs
            MissingEnvironmentError: Missing object storage informations
            StorageError: Storage read issue

        Returns:
            str: data, as binary string, None if no data
        """

        level_object = self.get_level(level)

        if level_object is None:
            raise Exception(f"No level {level} in the pyramid")

        if level_object.slab_width == 1 and level_object.slab_height == 1:
            raise NotImplementedError("One-tile slab pyramid is not handled")

        if not level_object.is_in_limits(column, row):
            return None

        # Indices de la dalle
        slab_column = column // level_object.slab_width
        slab_row = row // level_object.slab_height

        # Indices de la tuile dans la dalle
        relative_tile_column = column % level_object.slab_width
        relative_tile_row = row % level_object.slab_height

        # Numéro de la tuile dans le header
        tile_index = relative_tile_row * level_object.slab_width + relative_tile_column

        # Calcul du chemin de la dalle contenant la tuile voulue
        slab_path = self.get_slab_path_from_infos(SlabType.DATA, level, slab_column, slab_row)

        # Récupération des offset et tailles des tuiles dans la dalle
        # Une dalle ROK4 a une en-tête fixe de 2048 octets,
        # puis sont stockés les offsets (chacun sur 4 octets)
        # puis les tailles (chacune sur 4 octets)
        try:
            binary_index = get_data_binary(
                slab_path,
                (
                    ROK4_IMAGE_HEADER_SIZE,
                    2 * 4 * level_object.slab_width * level_object.slab_height,
                ),
            )
        except FileNotFoundError:
            # L'absence de la dalle est gérée comme simplement une absence de données
            return None

        offsets = numpy.frombuffer(
            binary_index,
            dtype=numpy.dtype("uint32"),
            count=level_object.slab_width * level_object.slab_height,
        )
        sizes = numpy.frombuffer(
            binary_index,
            dtype=numpy.dtype("uint32"),
            offset=4 * level_object.slab_width * level_object.slab_height,
            count=level_object.slab_width * level_object.slab_height,
        )

        if sizes[tile_index] == 0:
            return None

        return get_data_binary(slab_path, (offsets[tile_index], sizes[tile_index]))

    def get_tile_data_raster(self, level: str, column: int, row: int) -> numpy.ndarray:
        """Get a raster pyramid's tile as 3-dimension numpy ndarray

        First dimension is the row, second one is column, third one is band.

        Args:
            level (str): Tile's level
            column (int): Tile's column
            row (int): Tile's row

        Limitations:
            Packbits (pyramid formats TIFF_PKB_FLOAT32 and TIFF_PKB_UINT8) and LZW (pyramid formats TIFF_LZW_FLOAT32 and TIFF_LZW_UINT8) compressions are not handled.

        Raises:
            Exception: Cannot get raster data for a vector pyramid
            Exception: Level not found in the pyramid
            NotImplementedError: Pyramid owns one-tile slabs
            NotImplementedError: Raster pyramid format not handled
            MissingEnvironmentError: Missing object storage informations
            StorageError: Storage read issue
            FormatError: Cannot decode tile

        Examples:

            FILE stored DTM (raster) pyramid, to get the altitude value at a point in the best level

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("/data/pyramids/RGEALTI.json")
                    level, col, row, pcol, prow = pyramid.get_tile_indices(44, 5, srs = "EPSG:4326")
                    data = pyramid.get_tile_data_raster(level, col, row)

                    if data is None:
                        print("No data")
                    else:
                        print(data[prow][pcol])

                except Exception as e:
                    print("Cannot get a pyramid's pixel value : {e}")

        Returns:
            str: data, as numpy array, None if no data
        """

        if self.type == PyramidType.VECTOR:
            raise Exception("Cannot get tile as raster data : it's a vector pyramid")

        binary_tile = self.get_tile_data_binary(level, column, row)

        if binary_tile is None:
            return None

        level_object = self.get_level(level)

        if self.__format == "TIFF_JPG_UINT8" or self.__format == "TIFF_JPG90_UINT8":
            try:
                img = Image.open(io.BytesIO(binary_tile))
            except Exception as e:
                raise FormatError("JPEG", "binary tile", e)

            data = numpy.asarray(img)
            data.shape = (
                level_object.tile_matrix.tile_size[0],
                level_object.tile_matrix.tile_size[1],
                self.__raster_specifications["channels"],
            )

        elif self.__format == "TIFF_RAW_UINT8":
            data = numpy.frombuffer(binary_tile, dtype=numpy.dtype("uint8"))
            data.shape = (
                level_object.tile_matrix.tile_size[0],
                level_object.tile_matrix.tile_size[1],
                self.__raster_specifications["channels"],
            )

        elif self.__format == "TIFF_PNG_UINT8":
            try:
                img = Image.open(io.BytesIO(binary_tile))
            except Exception as e:
                raise FormatError("PNG", "binary tile", e)

            data = numpy.asarray(img)
            data.shape = (
                level_object.tile_matrix.tile_size[0],
                level_object.tile_matrix.tile_size[1],
                self.__raster_specifications["channels"],
            )

        elif self.__format == "TIFF_ZIP_UINT8":
            try:
                data = numpy.frombuffer(zlib.decompress(binary_tile), dtype=numpy.dtype("uint8"))
            except Exception as e:
                raise FormatError("ZIP", "binary tile", e)

            data.shape = (
                level_object.tile_matrix.tile_size[0],
                level_object.tile_matrix.tile_size[1],
                self.__raster_specifications["channels"],
            )

        elif self.__format == "TIFF_ZIP_FLOAT32":
            try:
                data = numpy.frombuffer(zlib.decompress(binary_tile), dtype=numpy.dtype("float32"))
            except Exception as e:
                raise FormatError("ZIP", "binary tile", e)

            data.shape = (
                level_object.tile_matrix.tile_size[0],
                level_object.tile_matrix.tile_size[1],
                self.__raster_specifications["channels"],
            )

        elif self.__format == "TIFF_RAW_FLOAT32":
            data = numpy.frombuffer(binary_tile, dtype=numpy.dtype("float32"))
            data.shape = (
                level_object.tile_matrix.tile_size[0],
                level_object.tile_matrix.tile_size[1],
                self.__raster_specifications["channels"],
            )

        else:
            raise NotImplementedError(f"Cannot get tile as raster data for format {self.__format}")

        return data

    def get_tile_data_vector(self, level: str, column: int, row: int) -> Dict:
        """Get a vector pyramid's tile as GeoJSON dictionnary

        Args:
            level (str): Tile's level
            column (int): Tile's column
            row (int): Tile's row

        Raises:
            Exception: Cannot get vector data for a raster pyramid
            Exception: Level not found in the pyramid
            NotImplementedError: Pyramid owns one-tile slabs
            NotImplementedError: Vector pyramid format not handled
            MissingEnvironmentError: Missing object storage informations
            StorageError: Storage read issue
            FormatError: Cannot decode tile

        Examples:

            S3 stored vector pyramid, to print a tile as GeoJSON

                from rok4.pyramid import Pyramid

                import json

                try:
                    pyramid = Pyramid.from_descriptor("s3://pyramids/vectors/BDTOPO.json")
                    level, col, row, pcol, prow = pyramid.get_tile_indices(40.325, 3.123, srs = "EPSG:4326")
                    data = pyramid.get_tile_data_vector(level, col, row)

                    if data is None:
                        print("No data")
                    else:
                        print(json.dumps(data))

                except Exception as e:
                    print("Cannot print a vector pyramid's tile as GeoJSON : {e}")

        Returns:
            str: data, as GeoJSON dictionnary. None if no data
        """

        if self.type == PyramidType.RASTER:
            raise Exception("Cannot get tile as vector data : it's a raster pyramid")

        binary_tile = self.get_tile_data_binary(level, column, row)

        if binary_tile is None:
            return None

        self.get_level(level)

        if self.__format == "TIFF_PBF_MVT":
            try:
                data = mapbox_vector_tile.decode(binary_tile)
            except Exception as e:
                raise FormatError("PBF (MVT)", "binary tile", e)
        else:
            raise NotImplementedError(f"Cannot get tile as vector data for format {self.__format}")

        return data

    def get_tile_indices(
        self, x: float, y: float, level: str = None, **kwargs
    ) -> Tuple[str, int, int, int, int]:
        """Get pyramid's tile and pixel indices from point's coordinates

        Used coordinates system have to be the pyramid one. If EPSG:4326, x is latitude and y longitude.

        Args:
            x (float): point's x
            y (float): point's y
            level (str, optional): Pyramid's level to take into account, the bottom one if None . Defaults to None.
            **srs (string): spatial reference system of provided coordinates, with authority and code (same as the pyramid's one if not provided)

        Raises:
            Exception: Cannot find level to calculate indices
            RuntimeError: Provided SRS is invalid for OSR

        Examples:

            FILE stored DTM (raster) pyramid, to get the altitude value at a point in the best level

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("/data/pyramids/RGEALTI.json")
                    level, col, row, pcol, prow = pyramid.get_tile_indices(44, 5, srs = "EPSG:4326")
                    data = pyramid.get_tile_data_raster(level, col, row)

                    if data is None:
                        print("No data")
                    else:
                        print(data[prow][pcol])

                except Exception as e:
                    print("Cannot get a pyramid's pixel value : {e}")

        Returns:
            Tuple[str, int, int, int, int]: Level identifier, tile's column, tile's row, pixel's (in the tile) column, pixel's row
        """

        level_object = self.bottom_level
        if level is not None:
            level_object = self.get_level(level)

        if level_object is None:
            raise Exception("Cannot found the level to calculate indices")

        if (
            "srs" in kwargs
            and kwargs["srs"] is not None
            and kwargs["srs"].upper() != self.__tms.srs.upper()
        ):
            sr = srs_to_spatialreference(kwargs["srs"])
            x, y = reproject_point((x, y), sr, self.__tms.sr)

        return (level_object.id,) + level_object.tile_matrix.point_to_indices(x, y)

    @property
    def size(self) -> int:
        """Get the size of the pyramid

        Examples:

                from rok4.pyramid import Pyramid

                try:
                    pyramid = Pyramid.from_descriptor("s3://bucket_name/path/to/descriptor.json")
                    size = pyramid.size()

                except Exception as e:
                    print("Cannot load the pyramid from its descriptor and get his size")

        Returns:
            int: size of the pyramid
        """

        if not hasattr(self, "_Pyramid__size"):
            self.__size = size_path(
                get_path_from_infos(self.__storage["type"], self.__storage["root"], self.__name)
            )

        return self.__size



import os
from unittest import mock
from unittest.mock import *

import botocore.exceptions
import pytest

from rok4.enums import StorageType
from rok4.exceptions import *
from rok4.storage import *


@mock.patch.dict(os.environ, {}, clear=True)
@patch("builtins.open", new_callable=mock_open, read_data=b"data")
def test_hash_file_ok(mock_file):
    try:
        md5 = hash_file("/path/to/file.ext")
        mock_file.assert_called_with("/path/to/file.ext", "rb")
        assert md5 == "8d777f385d3dfec8815d20f7496026dc"
    except Exception as exc:
        assert False, f"FILE md5 sum raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
def test_get_infos_from_path():
    assert (StorageType.S3, "toto/titi", "toto", "titi") == get_infos_from_path("s3://toto/titi")
    assert (
        StorageType.FILE,
        "/toto/titi/tutu.json",
        "/toto/titi",
        "tutu.json",
    ) == get_infos_from_path("file:///toto/titi/tutu.json")
    assert (StorageType.CEPH, "toto/titi/tutu", "toto", "titi/tutu") == get_infos_from_path(
        "ceph://toto/titi/tutu"
    )
    assert (StorageType.FILE, "wrong://toto/titi", "wrong://toto", "titi") == get_infos_from_path(
        "wrong://toto/titi"
    )


@mock.patch.dict(os.environ, {}, clear=True)
def test_get_path_from_infos():
    assert get_path_from_infos(StorageType.S3, "toto", "toto/titi") == "s3://toto/toto/titi"
    assert (
        get_path_from_infos(StorageType.FILE, "/toto/titi", "tutu.json")
        == "file:///toto/titi/tutu.json"
    )
    assert get_path_from_infos(StorageType.CEPH, "toto", "titi/tutu") == "ceph://toto/titi/tutu"


############ get_data_str


@mock.patch.dict(os.environ, {}, clear=True)
def test_s3_missing_env():
    with pytest.raises(MissingEnvironmentError):
        get_data_str("s3://bucket/path/to/object")


@mock.patch.dict(os.environ, {}, clear=True)
def test_ceph_missing_env():
    with pytest.raises(MissingEnvironmentError):
        get_data_str("ceph://bucket/path/to/object")


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "a,b", "ROK4_S3_SECRETKEY": "b,c", "ROK4_S3_KEY": "c,d,e"},
    clear=True,
)
def test_s3_invalid_envs():
    with pytest.raises(StorageError):
        get_data_str("s3://bucket/path/to/object")


@mock.patch.dict(
    os.environ, {"ROK4_S3_URL": "a", "ROK4_S3_SECRETKEY": "b", "ROK4_S3_KEY": "c"}, clear=True
)
@mock.patch("rok4.storage.boto3.client")
def test_s3_invalid_endpoint(mocked_s3_client):
    MagicMock()
    mocked_s3_client.side_effect = Exception("Invalid URL")
    with pytest.raises(StorageError):
        get_data_str("s3://bucket/path/to/object")


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("builtins.open", side_effect=FileNotFoundError("not_found"))
def test_file_read_error(mock_file):
    with pytest.raises(FileNotFoundError):
        get_data_str("file:///path/to/file.ext")

    mock_file.assert_called_with("/path/to/file.ext", "rb")


@mock.patch.dict(os.environ, {}, clear=True)
@patch("builtins.open", new_callable=mock_open, read_data=b"data")
def test_file_read_ok(mock_file):
    try:
        data = get_data_str("file:///path/to/file.ext")
        mock_file.assert_called_with("/path/to/file.ext", "rb")
        assert data == "data"
    except Exception as exc:
        assert False, f"FILE read raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_s3_read_nok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.get_object.side_effect = Exception("Bucket or object not found")
    mocked_s3_client.return_value = s3_instance
    with pytest.raises(StorageError):
        get_data_str("s3://bucket/path/to/object")


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_s3_read_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_body = MagicMock()
    s3_body.read.return_value = b"data"
    s3_instance.get_object.return_value = {"Body": s3_body}
    mocked_s3_client.return_value = s3_instance

    try:
        data = get_data_str("s3://bucket/path/to/object")
        assert data == "data"
    except Exception as exc:
        assert False, f"S3 read raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
def test_ceph_read_ok(mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.stat.return_value = (4, "date")
    ioctx_instance.read.return_value = b"data"
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        data = get_data_str("ceph://pool/path/to/object")
        assert data == "data"
    except Exception as exc:
        assert False, f"CEPH read raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("requests.get", side_effect={"status_code": 404})
def test_http_read_error(mock_http):
    with pytest.raises(StorageError):
        requests_instance = MagicMock()
        requests_instance.content = "NULL"
        requests_instance.status_code = 404
        mock_http.return_value = requests_instance
        get_data_str("http://path/to/file.ext")

    mock_http.assert_called_with("http://path/to/file.ext", stream=True)


@mock.patch.dict(os.environ, {}, clear=True)
def test_http_read_range_error():
    with pytest.raises(NotImplementedError):
        get_data_binary("http://path/to/file.ext", (0, 100))


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("requests.get")
def test_http_read_ok(mock_http):
    try:
        requests_instance = MagicMock()
        requests_instance.content = b"data"
        mock_http.return_value = requests_instance

        data = get_data_str("http://path/to/file.ext")
        mock_http.assert_called_with("http://path/to/file.ext", stream=True)
        assert data == "data"
    except Exception as exc:
        assert False, f"HTTP read raises an exception: {exc}"


############ put_data_str


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_s3_write_nok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.put_object.side_effect = Exception("Cannot write S3 object")
    mocked_s3_client.return_value = s3_instance

    with pytest.raises(StorageError):
        put_data_str("data", "s3://bucket/path/to/object")


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_s3_write_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.put_object.return_value = None
    mocked_s3_client.return_value = s3_instance
    try:
        put_data_str("data", "s3://bucket/path/to/object")
    except Exception as exc:
        assert False, f"S3 write raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
def test_ceph_write_ok(mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.write_full.return_value = None
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        put_data_str("data", "ceph://pool/path/to/object")
    except Exception as exc:
        assert False, f"CEPH write raises an exception: {exc}"


############ copy


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("os.makedirs", return_value=None)
@mock.patch("rok4.storage.copyfile", return_value=None)
@mock.patch("rok4.storage.hash_file", return_value="toto")
def test_copy_file_file_ok(mock_hash_file, mock_copyfile, mock_makedirs):
    try:
        copy("file:///path/to/source.ext", "file:///path/to/destination.ext", "toto")
        mock_copyfile.assert_called_once_with("/path/to/source.ext", "/path/to/destination.ext")
        mock_hash_file.assert_called_once_with("/path/to/destination.ext")
        mock_makedirs.assert_called_once_with("/path/to", exist_ok=True)
    except Exception as exc:
        assert False, f"FILE -> FILE copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
@mock.patch("os.makedirs", return_value=None)
@mock.patch("rok4.storage.hash_file", return_value="toto")
def test_copy_s3_file_ok(mock_hash_file, mock_makedirs, mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.download_file.return_value = None
    mocked_s3_client.return_value = s3_instance

    try:
        copy("s3://bucket/source.ext", "file:///path/to/destination.ext", "toto")
        mock_hash_file.assert_called_once_with("/path/to/destination.ext")
        mock_makedirs.assert_called_once_with("/path/to", exist_ok=True)
    except Exception as exc:
        assert False, f"S3 -> FILE copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
@mock.patch("os.makedirs", return_value=None)
@mock.patch("rok4.storage.hash_file", return_value="toto")
def test_copy_s3_file_nok(mock_hash_file, mock_makedirs, mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.download_file.side_effect = Exception("Cannot download S3 object")
    mocked_s3_client.return_value = s3_instance

    with pytest.raises(StorageError):
        copy("s3://bucket/source.ext", "file:///path/to/destination.ext", "toto")
        mock_makedirs.assert_called_once_with("/path/to", exist_ok=True)


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_copy_file_s3_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.upload_file.return_value = None
    s3_instance.head_object.return_value = {"ETag": "toto"}
    mocked_s3_client.return_value = s3_instance

    try:
        copy("file:///path/to/source.ext", "s3://bucket/destination.ext", "toto")
    except Exception as exc:
        assert False, f"FILE -> S3 copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_copy_s3_s3_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.copy.return_value = None
    s3_instance.head_object.return_value = {"ETag": "toto"}
    mocked_s3_client.return_value = s3_instance

    try:
        copy("s3://bucket/source.ext", "s3://bucket/destination.ext", "toto")
    except Exception as exc:
        assert False, f"S3 -> S3 copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_copy_s3_s3_intercluster_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.copy.return_value = None
    s3_instance.head_object.return_value = {"ETag": "toto"}
    mocked_s3_client.return_value = s3_instance

    try:
        copy("s3://bucket@a/source.ext", "s3://bucket@b/destination.ext", "toto")
    except Exception as exc:
        assert False, f"S3 -> S3 inter cluster copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_copy_s3_s3_intercluster_nok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.copy.return_value = None
    s3_instance.head_object.return_value = {"ETag": "toto"}
    mocked_s3_client.return_value = s3_instance

    with pytest.raises(StorageError):
        copy("s3://bucket@a/source.ext", "s3://bucket@c/destination.ext", "toto")


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
@mock.patch("os.makedirs", return_value=None)
@patch("builtins.open", new_callable=mock_open)
def test_copy_ceph_file_ok(mock_file, mock_makedirs, mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.read.return_value = b"data"
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        copy(
            "ceph://pool/source.ext",
            "file:///path/to/destination.ext",
            "8d777f385d3dfec8815d20f7496026dc",
        )
        mock_makedirs.assert_called_once_with("/path/to", exist_ok=True)
    except Exception as exc:
        assert False, f"CEPH -> FILE copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
@patch("builtins.open", new_callable=mock_open, read_data=b"data")
def test_copy_file_ceph_ok(mock_file, mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.write.return_value = None
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        copy(
            "file:///path/to/source.ext",
            "ceph://pool/destination.ext",
            "8d777f385d3dfec8815d20f7496026dc",
        )
    except Exception as exc:
        assert False, f"FILE -> CEPH copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
@patch("builtins.open", new_callable=mock_open, read_data=b"data")
def test_copy_ceph_ceph_ok(mock_file, mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.read.return_value = b"data"
    ioctx_instance.write.return_value = None
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        copy(
            "ceph://pool1/source.ext",
            "ceph://pool2/destination.ext",
            "8d777f385d3dfec8815d20f7496026dc",
        )
    except Exception as exc:
        assert False, f"CEPH -> CEPH copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {
        "ROK4_CEPH_CONFFILE": "a",
        "ROK4_CEPH_CLUSTERNAME": "b",
        "ROK4_CEPH_USERNAME": "c",
        "ROK4_S3_URL": "https://a,https://b",
        "ROK4_S3_SECRETKEY": "a,b",
        "ROK4_S3_KEY": "a,b",
    },
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
@mock.patch("rok4.storage.boto3.client")
@patch("builtins.open", new_callable=mock_open, read_data=b"data")
def test_copy_ceph_s3_ok(mock_file, mocked_s3_client, mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.read.return_value = b"data"
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.upload_file.return_value = None
    s3_instance.head_object.return_value = {"ETag": "8d777f385d3dfec8815d20f7496026dc"}
    mocked_s3_client.return_value = s3_instance

    try:
        copy(
            "ceph://pool1/source.ext",
            "s3://bucket/destination.ext",
            "8d777f385d3dfec8815d20f7496026dc",
        )
    except Exception as exc:
        assert False, f"CEPH -> S3 copy raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("requests.get")
@patch("builtins.open", new_callable=mock_open)
def test_copy_http_file_ok(mock_open, mock_requests):
    try:
        http_instance = MagicMock()
        http_instance.iter_content.return_value = ["data", "data2"]
        mock_requests.return_value = http_instance

        copy("http://path/to/source.ext", "file:///path/to/destination.ext")
        mock_requests.assert_called_once_with("http://path/to/source.ext", stream=True)
        mock_open.assert_called_once_with("/path/to/destination.ext", "wb")
    except Exception as exc:
        assert False, f"HTTP -> FILE copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
@mock.patch("requests.get")
def test_copy_http_ceph_ok(mock_requests, mocked_rados_client):
    try:
        http_instance = MagicMock()
        http_instance.iter_content.return_value = ["data", "data2"]
        mock_requests.return_value = http_instance

        disconnect_ceph_clients()
        ioctx_instance = MagicMock()
        ioctx_instance.write.return_value = None
        ceph_instance = MagicMock()
        ceph_instance.open_ioctx.return_value = ioctx_instance
        mocked_rados_client.return_value = ceph_instance

        copy("http://path/to/source.ext", "ceph://pool1/source.ext")
        mock_requests.assert_called_once_with("http://path/to/source.ext", stream=True)
    except Exception as exc:
        assert False, f"HTTP -> CEPH copy raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
@mock.patch("requests.get")
@patch("tempfile.NamedTemporaryFile", new_callable=mock_open)
@mock.patch("os.remove")
def test_copy_http_s3_ok(mock_remove, mock_tempfile, mock_requests, mocked_s3_client):
    try:
        http_instance = MagicMock()
        http_instance.iter_content.return_value = ["data", "data2"]
        mock_requests.return_value = http_instance

        disconnect_s3_clients()
        s3_instance = MagicMock()
        s3_instance.upload_file.return_value = None
        s3_instance.head_object.return_value = {"ETag": "8d777f385d3dfec8815d20f7496026dc"}
        mocked_s3_client.return_value = s3_instance

        copy("http://path/to/source.ext", "s3://bucket/destination.ext")
        mock_requests.assert_called_once_with("http://path/to/source.ext", stream=True)
        mock_tempfile.assert_called_once_with("w+b", delete=False)
    except Exception as exc:
        assert False, f"HTTP -> CEPH copy raises an exception: {exc}"


############ link


def test_link_type_nok():
    with pytest.raises(StorageError):
        link("ceph://pool1/target.ext", "file:///path/to/link.ext")


def test_link_hard_nok():
    with pytest.raises(StorageError):
        link("ceph://pool1/source.ext", "ceph://pool2/destination.ext", True)


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("os.symlink", return_value=None)
def test_link_file_ok(mock_link):
    try:
        link("file:///path/to/target.ext", "file:///path/to/link.ext")
        mock_link.assert_called_once_with("/path/to/target.ext", "/path/to/link.ext")
    except Exception as exc:
        assert False, f"FILE link raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("os.link", return_value=None)
def test_hlink_file_ok(mock_link):
    try:
        link("file:///path/to/target.ext", "file:///path/to/link.ext", True)
        mock_link.assert_called_once_with("/path/to/target.ext", "/path/to/link.ext")
    except Exception as exc:
        assert False, f"FILE hard link raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
def test_link_ceph_ok(mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.write.return_value = None
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        link("ceph://pool1/target.ext", "ceph://pool2/link.ext")
    except Exception as exc:
        assert False, f"CEPH link raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_link_s3_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.put_object.return_value = None
    mocked_s3_client.return_value = s3_instance

    try:
        link("s3://bucket1/target.ext", "s3://bucket2/link.ext")
    except Exception as exc:
        assert False, f"S3 link raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_link_s3_nok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.put_object.return_value = None
    mocked_s3_client.return_value = s3_instance

    with pytest.raises(StorageError):
        link("s3://bucket1@a/target.ext", "s3://bucket2@b/link.ext")


############ get_size


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("os.stat")
def test_size_file_ok(mock_stat):
    mock_stat.return_value.st_size = 12
    try:
        size = get_size("file:///path/to/file.ext")
        assert size == 12
    except Exception as exc:
        assert False, f"FILE size raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
def test_size_ceph_ok(mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.stat.return_value = (12, "date")
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        size = get_size("ceph://pool/object.ext")
        assert size == 12
    except Exception as exc:
        assert False, f"CEPH size raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_size_s3_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.head_object.return_value = {"ContentLength": 12}
    mocked_s3_client.return_value = s3_instance

    try:
        size = get_size("s3://bucket/object.ext")
        assert size == 12
    except Exception as exc:
        assert False, f"S3 size raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("requests.get")
def test_size_http_ok(mock_requests):
    http_instance = MagicMock()
    http_instance.headers = {"content-length": 12}
    mock_requests.return_value = http_instance

    try:
        size = get_size("http://path/to/file.ext")
        assert size == 12
    except Exception as exc:
        assert False, f"HTTP size raises an exception: {exc}"


############ exists


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("os.path.exists", return_value=True)
def test_exists_file_ok(mock_exists):
    try:
        assert exists("file:///path/to/file.ext")
    except Exception as exc:
        assert False, f"FILE exists raises an exception: {exc}"

    mock_exists.return_value = False
    try:
        assert not exists("file:///path/to/file.ext")
    except Exception as exc:
        assert False, f"FILE not exists raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
def test_exists_ceph_ok(mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.stat.return_value = None
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        assert exists("ceph://pool/object.ext")
    except Exception as exc:
        assert False, f"CEPH exists raises an exception: {exc}"

    ioctx_instance.stat.side_effect = rados.ObjectNotFound("error")
    try:
        assert not exists("ceph://pool/object.ext")
    except Exception as exc:
        assert False, f"CEPH not exists raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_exists_s3_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.head_object.return_value = None
    mocked_s3_client.return_value = s3_instance

    try:
        assert exists("s3://bucket/object.ext")
    except Exception as exc:
        assert False, f"S3 exists raises an exception: {exc}"

    s3_instance.head_object.side_effect = botocore.exceptions.ClientError(
        operation_name="InvalidKeyPair.Duplicate", error_response={"Error": {"Code": "404"}}
    )
    try:
        assert not exists("s3://bucket/object.ext")
    except Exception as exc:
        assert False, f"CEPH not exists raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("requests.get")
def test_exists_http_ok(mock_requests):
    http_instance = MagicMock()
    http_instance.status_code = 200
    mock_requests.return_value = http_instance

    try:
        assert exists("http://path/to/file.ext")
    except Exception as exc:
        assert False, f"HTTP exists raises an exception: {exc}"

    http_instance.status_code = 404
    mock_requests.return_value = http_instance

    try:
        assert not exists("http://path/to/file.ext")
    except Exception as exc:
        assert False, f"HTTP exists raises an exception: {exc}"


############ remove


@mock.patch.dict(os.environ, {}, clear=True)
@mock.patch("os.remove")
def test_remove_file_ok(mock_remove):
    mock_remove.return_value = None
    try:
        remove("file:///path/to/file.ext")
    except Exception as exc:
        assert False, f"FILE deletion raises an exception: {exc}"

    mock_remove.side_effect = FileNotFoundError("error")
    try:
        remove("file:///path/to/file.ext")
    except Exception as exc:
        assert False, f"FILE deletion (not found) raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
@mock.patch("rok4.storage.rados.Rados")
def test_remove_ceph_ok(mocked_rados_client):
    disconnect_ceph_clients()
    ioctx_instance = MagicMock()
    ioctx_instance.remove_object.return_value = None
    ceph_instance = MagicMock()
    ceph_instance.open_ioctx.return_value = ioctx_instance
    mocked_rados_client.return_value = ceph_instance

    try:
        remove("ceph://pool/object.ext")
    except Exception as exc:
        assert False, f"CEPH deletion raises an exception: {exc}"

    ioctx_instance.stat.side_effect = rados.ObjectNotFound("error")
    try:
        Exception("ceph://pool/object.ext")
    except Exception as exc:
        assert False, f"CEPH deletion (not found) raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_remove_s3_ok(mocked_s3_client):
    disconnect_s3_clients()
    s3_instance = MagicMock()
    s3_instance.delete_object.return_value = None
    mocked_s3_client.return_value = s3_instance

    try:
        remove("s3://bucket/object.ext")
    except Exception as exc:
        assert False, f"S3 deletion raises an exception: {exc}"


############ get_osgeo_path


@mock.patch.dict(os.environ, {}, clear=True)
def test_get_osgeo_path_file_ok():
    try:
        path = get_osgeo_path("file:///path/to/file.ext")
        assert path == "/path/to/file.ext"
    except Exception as exc:
        assert False, f"FILE osgeo path raises an exception: {exc}"


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
def test_get_osgeo_path_s3_ok():
    disconnect_s3_clients()

    try:
        path = get_osgeo_path("s3://bucket@b/to/object.ext")
        assert path == "/vsis3/bucket/to/object.ext"
    except Exception as exc:
        assert False, f"S3 osgeo path raises an exception: {exc}"


@mock.patch.dict(os.environ, {}, clear=True)
def test_get_osgeo_path_nok():
    with pytest.raises(NotImplementedError):
        get_osgeo_path("ceph://pool/data.ext")


############ size_path
def test_size_path_file_ok():
    try:
        size = size_path("file://tests/fixtures/TIFF_PBF_MVT")
        assert size == 124838
    except Exception as exc:
        assert False, f"FILE size of the path raises an exception: {exc}"


def test_size_file_nok():
    with pytest.raises(StorageError):
        size_path("file://tests/fixtures/TIFF_PBF_M")


@mock.patch.dict(
    os.environ,
    {"ROK4_CEPH_CONFFILE": "a", "ROK4_CEPH_CLUSTERNAME": "b", "ROK4_CEPH_USERNAME": "c"},
    clear=True,
)
def test_size_path_ceph_nok():
    with pytest.raises(NotImplementedError):
        size_path("ceph://pool/path")


@mock.patch.dict(
    os.environ,
    {"ROK4_S3_URL": "https://a,https://b", "ROK4_S3_SECRETKEY": "a,b", "ROK4_S3_KEY": "a,b"},
    clear=True,
)
@mock.patch("rok4.storage.boto3.client")
def test_size_path_s3_ok(mocked_s3_client):
    disconnect_s3_clients()
    pages = [{"Contents": [{"Size": 10}, {"Size": 20}]}, {"Contents": [{"Size": 50}]}]
    paginator = MagicMock()
    paginator.paginate.return_value = pages
    client = MagicMock()
    client.get_paginator.return_value = paginator
    mocked_s3_client.return_value = client

    try:
        size = size_path("s3://bucket/path")
        assert size == 80
    except Exception as exc:
        assert False, f"S3 size of the path raises an exception: {exc}"



import math
import random
from unittest import mock
from unittest.mock import *

import pytest
from osgeo import gdal, osr

from rok4.exceptions import *
from rok4.utils import *


def test_srs_to_spatialreference_ignf_ok():
    try:
        srs_to_spatialreference("IGNF:LAMB93")
        srs_to_spatialreference("ignf:lamb93")
    except Exception as exc:
        assert False, f"SpatialReference creation raises an exception: {exc}"


def test_srs_to_spatialreference_epsg_ok():
    try:
        srs_to_spatialreference("EPSG:3857")
        srs_to_spatialreference("epsg:3857")
    except Exception as exc:
        assert False, f"SpatialReference creation raises an exception: {exc}"


def test_srs_to_spatialreference_ignf_nok():
    with pytest.raises(Exception):
        srs_to_spatialreference("IGNF:TOTO")


def test_srs_to_spatialreference_epsg_nok():
    with pytest.raises(Exception):
        srs_to_spatialreference("EPSG:123456")


def test_bbox_to_geometry_ok():
    try:
        geom = bbox_to_geometry((0, 0, 5, 10))
        assert geom.Area() == 50
    except Exception as exc:
        assert False, f"Geometry creation from bbox raises an exception: {exc}"


def test_reproject_bbox_ok():
    try:
        bbox = reproject_bbox((-90, -180, 90, 180), "EPSG:4326", "EPSG:3857")
        assert bbox[0] == -20037508.342789244
        bbox = reproject_bbox((43, 3, 44, 4), "EPSG:4326", "IGNF:WGS84G")
        assert bbox[0] == 3
    except Exception as exc:
        assert False, f"Bbox reprojection raises an exception: {exc}"


def test_reproject_point_ok():
    try:
        sr_4326 = srs_to_spatialreference("EPSG:4326")
        sr_3857 = srs_to_spatialreference("EPSG:3857")
        sr_ignf = srs_to_spatialreference("IGNF:WGS84G")
        x, y = reproject_point((43, 3), sr_4326, sr_3857)

        assert math.isclose(x, 333958.4723798207, rel_tol=1e-5)
        assert math.isclose(y, 5311971.846945471, rel_tol=1e-5)

        x, y = reproject_point((43, 3), sr_4326, sr_ignf)
        assert (x, y) == (3, 43)

        bbox = reproject_bbox((43, 3, 44, 4), "EPSG:4326", "IGNF:WGS84G")
        assert bbox[0] == 3
    except Exception as exc:
        assert False, f"Bbox reprojection raises an exception: {exc}"


# Tests for the rok4.utils.compute_bbox function.


def test_compute_bbox_epsg_3857_ok():
    try:
        mocked_datasource = MagicMock(gdal.Dataset)
        random.seed()

        image_size = (random.randint(1, 1000), random.randint(1, 1000))
        mocked_datasource.RasterXSize = image_size[0]
        mocked_datasource.RasterYSize = image_size[1]

        transform_tuple = (
            random.uniform(-10000000.0, 10000000.0),
            random.uniform(-10000, 10000),
            random.uniform(-10000, 10000),
            random.uniform(-10000000.0, 10000000.0),
            random.uniform(-10000, 10000),
            random.uniform(-10000, 10000),
        )
        mocked_datasource.GetGeoTransform = Mock(return_value=transform_tuple)
        mocked_spatial_ref = MagicMock(osr.SpatialReference)
        mocked_spatial_ref.GetDataAxisToSRSAxisMapping = Mock(return_value=[1, 2])
        mocked_datasource.GetSpatialRef = Mock(return_value=mocked_spatial_ref)

        x_range = (
            transform_tuple[0],
            transform_tuple[0]
            + image_size[0] * transform_tuple[1]
            + image_size[1] * transform_tuple[2],
        )
        y_range = (
            transform_tuple[3],
            transform_tuple[3]
            + image_size[0] * transform_tuple[4]
            + image_size[1] * transform_tuple[5],
        )

        expected = (min(x_range), min(y_range), max(x_range), max(y_range))
        result = compute_bbox(mocked_datasource)
        assert math.isclose(result[0], expected[0], rel_tol=1e-5)
        assert math.isclose(result[1], expected[1], rel_tol=1e-5)
        assert math.isclose(result[2], expected[2], rel_tol=1e-5)
        assert math.isclose(result[3], expected[3], rel_tol=1e-5)
        mocked_datasource.GetSpatialRef.assert_called_once()
        mocked_spatial_ref.GetDataAxisToSRSAxisMapping.assert_called()
    except Exception as exc:
        assert False, f"Bbox computation raises an exception: {exc}"


def test_compute_bbox_epsg_4326_ok():
    try:
        mocked_datasource = MagicMock(gdal.Dataset)
        random.seed()

        image_size = (random.randint(1, 1000), random.randint(1, 1000))
        mocked_datasource.RasterXSize = image_size[0]
        mocked_datasource.RasterYSize = image_size[1]

        transform_tuple = (
            random.uniform(-10000000.0, 10000000.0),
            random.uniform(-10000, 10000),
            random.uniform(-10000, 10000),
            random.uniform(-10000000.0, 10000000.0),
            random.uniform(-10000, 10000),
            random.uniform(-10000, 10000),
        )
        mocked_datasource.GetGeoTransform = Mock(return_value=transform_tuple)
        mocked_spatial_ref = MagicMock(osr.SpatialReference)
        mocked_spatial_ref.GetDataAxisToSRSAxisMapping = Mock(return_value=[2, 1])
        mocked_datasource.GetSpatialRef = Mock(return_value=mocked_spatial_ref)

        x_range = (
            transform_tuple[0],
            transform_tuple[0]
            + image_size[0] * transform_tuple[1]
            + image_size[1] * transform_tuple[2],
        )
        y_range = (
            transform_tuple[3],
            transform_tuple[3]
            + image_size[0] * transform_tuple[4]
            + image_size[1] * transform_tuple[5],
        )

        expected = (min(y_range), min(x_range), max(y_range), max(x_range))
        result = compute_bbox(mocked_datasource)
        assert math.isclose(result[0], expected[0], rel_tol=1e-5)
        assert math.isclose(result[1], expected[1], rel_tol=1e-5)
        assert math.isclose(result[2], expected[2], rel_tol=1e-5)
        assert math.isclose(result[3], expected[3], rel_tol=1e-5)
        mocked_datasource.GetSpatialRef.assert_called_once()
        mocked_spatial_ref.GetDataAxisToSRSAxisMapping.assert_called()
    except Exception as exc:
        assert False, f"Bbox computation raises an exception: {exc}"


def test_compute_bbox_no_srs_ok():
    try:
        mocked_datasource = MagicMock(gdal.Dataset)
        random.seed()

        image_size = (random.randint(1, 1000), random.randint(1, 1000))
        mocked_datasource.RasterXSize = image_size[0]
        mocked_datasource.RasterYSize = image_size[1]

        transform_tuple = (
            random.uniform(-10000000.0, 10000000.0),
            random.uniform(-10000, 10000),
            random.uniform(-10000, 10000),
            random.uniform(-10000000.0, 10000000.0),
            random.uniform(-10000, 10000),
            random.uniform(-10000, 10000),
        )
        mocked_datasource.GetGeoTransform = Mock(return_value=transform_tuple)
        mocked_datasource.GetSpatialRef = Mock(return_value=None)

        x_range = (
            transform_tuple[0],
            transform_tuple[0]
            + image_size[0] * transform_tuple[1]
            + image_size[1] * transform_tuple[2],
        )
        y_range = (
            transform_tuple[3],
            transform_tuple[3]
            + image_size[0] * transform_tuple[4]
            + image_size[1] * transform_tuple[5],
        )

        expected = (min(x_range), min(y_range), max(x_range), max(y_range))
        result = compute_bbox(mocked_datasource)
        assert math.isclose(result[0], expected[0], rel_tol=1e-5)
        assert math.isclose(result[1], expected[1], rel_tol=1e-5)
        assert math.isclose(result[2], expected[2], rel_tol=1e-5)
        assert math.isclose(result[3], expected[3], rel_tol=1e-5)
        mocked_datasource.GetSpatialRef.assert_called_once()
    except Exception as exc:
        assert False, f"Bbox computation raises an exception: {exc}"


# Tests for the rok4.utils.compute_format function.


@mock.patch("rok4.utils.gdal.Info")
@mock.patch("rok4.utils.gdal.GetColorInterpretationName", return_value="Palette")
@mock.patch("rok4.utils.gdal.GetDataTypeSize", return_value=8)
@mock.patch("rok4.utils.gdal.GetDataTypeName", return_value="Byte")
def test_compute_format_bit_ok(
    mocked_GetDataTypeName, mocked_GetDataTypeSize, mocked_GetColorInterpretationName, mocked_Info
):
    try:
        mocked_datasource = MagicMock(gdal.Dataset)

        mocked_datasource.RasterCount = 1
        mocked_Info.return_value = """Driver: GTiff/GeoTIFF
Size is 10000, 10000
Image Structure Metadata:
  COMPRESSION=PACKBITS
  INTERLEAVE=BAND
  MINISWHITE=YES
"""

        result = compute_format(mocked_datasource)
        assert result == ColorFormat.BIT
        mocked_GetDataTypeName.assert_called()
        mocked_GetDataTypeSize.assert_called()
        mocked_GetColorInterpretationName.assert_called()
        mocked_Info.assert_called()
    except Exception as exc:
        assert False, f"Color format computation raises an exception: {exc}"


@mock.patch("rok4.utils.gdal.Info")
@mock.patch("rok4.utils.gdal.GetColorInterpretationName")
@mock.patch("rok4.utils.gdal.GetDataTypeSize", return_value=8)
@mock.patch("rok4.utils.gdal.GetDataTypeName", return_value="Byte")
def test_compute_format_uint8_ok(
    mocked_GetDataTypeName, mocked_GetDataTypeSize, mocked_GetColorInterpretationName, mocked_Info
):
    try:
        mocked_datasource = MagicMock(gdal.Dataset)

        band_number = random.randint(1, 4)
        mocked_datasource.RasterCount = band_number
        band_name = None
        if band_number == 1 or band_number == 2:
            band_name = "Gray"
        elif band_number == 3 or band_number == 4:
            band_name = "Red"
        mocked_GetColorInterpretationName.return_value = band_name
        mocked_Info.return_value = """Driver: GTiff/GeoTIFF
Size is 10000, 10000
Metadata:
  AREA_OR_POINT=Area
Image Structure Metadata:
  INTERLEAVE=BAND
"""

        result = compute_format(mocked_datasource)

        assert result == ColorFormat.UINT8
        mocked_GetDataTypeName.assert_called()
        mocked_GetDataTypeSize.assert_called()
        mocked_GetColorInterpretationName.assert_called()
        mocked_Info.assert_called()
    except Exception as exc:
        assert False, f"Color format computation raises an exception: {exc}"


@mock.patch("rok4.utils.gdal.Info")
@mock.patch("rok4.utils.gdal.GetColorInterpretationName")
@mock.patch("rok4.utils.gdal.GetDataTypeSize", return_value=32)
@mock.patch("rok4.utils.gdal.GetDataTypeName", return_value="Float32")
def test_compute_format_float32_ok(
    mocked_GetDataTypeName, mocked_GetDataTypeSize, mocked_GetColorInterpretationName, mocked_Info
):
    try:
        mocked_datasource = MagicMock(gdal.Dataset)

        band_number = random.randint(1, 4)
        mocked_datasource.RasterCount = band_number
        band_name = None
        if band_number == 1 or band_number == 2:
            band_name = "Gray"
        elif band_number == 3 or band_number == 4:
            band_name = "Red"
        mocked_GetColorInterpretationName.return_value = band_name
        mocked_Info.return_value = """Driver: GTiff/GeoTIFF
Size is 10000, 10000
Metadata:
  AREA_OR_POINT=Area
Image Structure Metadata:
  INTERLEAVE=BAND
"""

        result = compute_format(mocked_datasource)

        assert result == ColorFormat.FLOAT32
        mocked_GetDataTypeName.assert_called()
        mocked_GetDataTypeSize.assert_called()
        mocked_GetColorInterpretationName.assert_called()
        mocked_Info.assert_called()
    except Exception as exc:
        assert False, f"Color format computation raises an exception: {exc}"


@mock.patch("rok4.utils.gdal.Info")
@mock.patch("rok4.utils.gdal.GetColorInterpretationName")
@mock.patch("rok4.utils.gdal.GetDataTypeSize", return_value=16)
@mock.patch("rok4.utils.gdal.GetDataTypeName", return_value="UInt16")
def test_compute_format_unsupported_nok(
    mocked_GetDataTypeName, mocked_GetDataTypeSize, mocked_GetColorInterpretationName, mocked_Info
):
    try:
        mocked_datasource = MagicMock(gdal.Dataset)

        band_number = random.randint(1, 4)
        mocked_datasource.RasterCount = band_number
        band_name = None
        if band_number == 1 or band_number == 2:
            band_name = "Gray"
        elif band_number == 3 or band_number == 4:
            band_name = "Red"
        mocked_GetColorInterpretationName.return_value = band_name
        mocked_Info.return_value = """Driver: GTiff/GeoTIFF
Size is 10000, 10000
Metadata:
  AREA_OR_POINT=Area
Image Structure Metadata:
  INTERLEAVE=BAND
"""

        with pytest.raises(Exception):
            compute_format(mocked_datasource)

        mocked_GetDataTypeName.assert_called()
        mocked_GetDataTypeSize.assert_called()
        mocked_GetColorInterpretationName.assert_called()
        mocked_Info.assert_called()
    except Exception as exc:
        assert False, f"Color format computation raises an exception: {exc}"


@mock.patch("rok4.utils.gdal.Info")
@mock.patch("rok4.utils.gdal.GetColorInterpretationName")
@mock.patch("rok4.utils.gdal.GetDataTypeSize", return_value=16)
@mock.patch("rok4.utils.gdal.GetDataTypeName", return_value="UInt16")
def test_compute_format_no_band_nok(
    mocked_GetDataTypeName, mocked_GetDataTypeSize, mocked_GetColorInterpretationName, mocked_Info
):
    try:
        mocked_datasource = MagicMock(gdal.Dataset)

        mocked_datasource.RasterCount = 0

        with pytest.raises(Exception):
            compute_format(mocked_datasource)

        mocked_GetDataTypeName.assert_not_called()
        mocked_GetDataTypeSize.assert_not_called()
        mocked_GetColorInterpretationName.assert_not_called()
        mocked_Info.assert_not_called()
    except Exception as exc:
        assert False, f"Color format computation raises an exception: {exc}"



import os
from unittest import mock
from unittest.mock import *

import pytest

from rok4.exceptions import *
from rok4.tile_matrix_set import TileMatrixSet


@mock.patch.dict(os.environ, {}, clear=True)
def test_missing_env():
    with pytest.raises(MissingEnvironmentError):
        TileMatrixSet("tms")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch("rok4.tile_matrix_set.get_data_str", side_effect=StorageError("FILE", "Not found"))
def test_wrong_file(mocked_get_data_str):
    with pytest.raises(StorageError):
        TileMatrixSet("tms")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='"crs":"EPSG:3857","orderedAxes":["X","Y"],"id":"PM"}',
)
def test_bad_json(mocked_get_data_str):
    with pytest.raises(FormatError):
        TileMatrixSet("tms")
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"tileMatrices":[{"id":"0","tileWidth":256,"scaleDenominator":559082264.028718,"matrixWidth":1,"cellSize":156543.033928041,"matrixHeight":1,"tileHeight":256,"pointOfOrigin":[-20037508.3427892,20037508.3427892]}],"crs":"EPSG:3857","orderedAxes":["X","Y"]}',
)
def test_missing_id(mocked_get_data_str):
    with pytest.raises(MissingAttributeError) as exc:
        TileMatrixSet("tms")
    assert str(exc.value) == "Missing attribute 'id' in 'file:///path/to/tms.json'"
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"tileMatrices":[{"id":"0","tileWidth":256,"scaleDenominator":559082264.028718,"matrixWidth":1,"cellSize":156543.033928041,"matrixHeight":1,"tileHeight":256,"pointOfOrigin":[-20037508.3427892,20037508.3427892]}],"orderedAxes":["X","Y"],"id":"PM"}',
)
def test_missing_crs(mocked_get_data_str):
    with pytest.raises(MissingAttributeError) as exc:
        TileMatrixSet("tms")
    assert str(exc.value) == "Missing attribute 'crs' in 'file:///path/to/tms.json'"
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"crs":"epsg:123456","orderedAxes":["X","Y"],"tileMatrices":[{"id":"0","tileWidth":256,"scaleDenominator":559082264.028718,"matrixWidth":1,"cellSize":156543.033928041,"matrixHeight":1,"tileHeight":256,"pointOfOrigin":[-20037508.3427892,20037508.3427892]}],"orderedAxes":["X","Y"],"id":"PM"}',
)
def test_wrong_crs(mocked_get_data_str):
    with pytest.raises(Exception) as exc:
        TileMatrixSet("tms")
    assert (
        str(exc.value)
        == "Wrong attribute 'crs' ('epsg:123456') in 'file:///path/to/tms.json', not recognize by OSR. Trace : PROJ: proj_create_from_database: crs not found"
    )
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"crs":"epsg:4326","orderedAxes":["Lat","Lon"],"tileMatrices":[{"id":"0","tileWidth":256,"scaleDenominator":559082264.028718,"matrixWidth":1,"cellSize":156543.033928041,"matrixHeight":1,"tileHeight":256,"pointOfOrigin":[-20037508.3427892,20037508.3427892]}],"id":"PM"}',
)
def test_wrong_axes_order(mocked_get_data_str):
    with pytest.raises(Exception) as exc:
        TileMatrixSet("tms")
    assert (
        str(exc.value)
        == "TMS 'file:///path/to/tms.json' own invalid axes order : only X/Y or Lon/Lat are handled"
    )
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"crs":"EPSG:3857","orderedAxes":["X","Y"],"id":"PM"}',
)
def test_missing_levels(mocked_get_data_str):
    with pytest.raises(MissingAttributeError) as exc:
        TileMatrixSet("tms")
    assert str(exc.value) == "Missing attribute 'tileMatrices' in 'file:///path/to/tms.json'"
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"tileMatrices":[],"crs":"EPSG:3857","orderedAxes":["X","Y"],"id":"PM"}',
)
def test_no_levels(mocked_get_data_str):
    with pytest.raises(Exception) as exc:
        TileMatrixSet("tms")
    assert str(exc.value) == "TMS 'file:///path/to/tms.json' has no level"
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"tileMatrices":[{"tileWidth":256,"scaleDenominator":559082264.028718,"matrixWidth":1,"cellSize":156543.033928041,"matrixHeight":1,"tileHeight":256,"pointOfOrigin":[-20037508.3427892,20037508.3427892]}],"orderedAxes":["X","Y"],"id":"PM","crs":"EPSG:3857"}',
)
def test_wrong_level(mocked_get_data_str):
    with pytest.raises(MissingAttributeError) as exc:
        TileMatrixSet("tms")
    assert str(exc.value) == "Missing attribute tileMatrices[].'id' in 'file:///path/to/tms.json'"
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"tileMatrices":[{"id":"level_0","tileWidth":256,"scaleDenominator":559082264.028718,"matrixWidth":1,"cellSize":156543.033928041,"matrixHeight":1,"tileHeight":256,"pointOfOrigin":[-20037508.3427892,20037508.3427892]}],"crs":"EPSG:3857","orderedAxes":["X","Y"],"id":"PM"}',
)
def test_wrong_level_id(mocked_get_data_str):
    with pytest.raises(Exception) as exc:
        TileMatrixSet("tms")

    assert (
        str(exc.value)
        == "TMS file:///path/to/tms.json owns a level whom id contains an underscore (level_0)"
    )
    mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"tileMatrices":[{"id":"0","tileWidth":256,"scaleDenominator":559082264.028718,"matrixWidth":1,"cellSize":156543.033928041,"matrixHeight":1,"tileHeight":256,"pointOfOrigin":[-20037508.3427892,20037508.3427892]}],"crs":"EPSG:3857","orderedAxes":["X","Y"],"id":"PM"}',
)
def test_ok(mocked_get_data_str):
    try:
        tms = TileMatrixSet("tms")
        assert tms.get_level("0") is not None
        assert tms.get_level("4") is None
        mocked_get_data_str.assert_called_once_with("file:///path/to/tms.json")
    except Exception as exc:
        assert False, f"'TileMatrixSet creation raises an exception: {exc}"


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"tileMatrices":[{"id":"17","cellSize":1.19432856695588,"matrixHeight":131072,"pointOfOrigin":[-20037508.3427892,20037508.3427892],"tileHeight":256,"tileWidth":256,"scaleDenominator":4265.45916769957,"matrixWidth":131072}],"crs":"EPSG:3857","orderedAxes":["X","Y"],"id":"PM"}',
)
def test_pm_conversions(mocked_get_data_str):
    try:
        tms = TileMatrixSet("tms")
        tm = tms.get_level("17")
        assert tm.x_to_column(670654.2832369965) == 67729
        assert tm.y_to_row(5980575.503117723) == 45975
        assert tm.tile_to_bbox(67728, 45975) == (
            670199.864004489,
            5980433.093032133,
            670505.6121176295,
            5980738.841145273,
        )
        assert tm.bbox_to_tiles(
            (670034.4267107458, 5980565.948489188, 670649.5059227281, 5980936.190344945)
        ) == (67727, 45974, 67729, 45975)
        assert tm.point_to_indices(670654.2832369965, 5980575.503117723) == (67729, 45975, 124, 136)
    except Exception as exc:
        assert False, f"'TileMatrixSet creation raises an exception: {exc}"


@mock.patch.dict(os.environ, {"ROK4_TMS_DIRECTORY": "file:///path/to"}, clear=True)
@mock.patch(
    "rok4.tile_matrix_set.get_data_str",
    return_value='{"crs":"EPSG:4326","tileMatrices":[{"tileWidth":256,"scaleDenominator":1066.36480348451,"matrixWidth":524288,"cellSize":2.68220901489258e-06,"matrixHeight":262144,"pointOfOrigin":[-180,90],"tileHeight":256,"id":"18"}],"orderedAxes":["Lon","Lat"],"id":"4326"}',
)
def test_4326_conversions(mocked_get_data_str):
    try:
        tms = TileMatrixSet("tms")
        tm = tms.get_level("18")
        assert tm.x_to_column(5) == 269425
        assert tm.y_to_row(45) == 65535
        assert tm.tile_to_bbox(269425, 65535) == (
            44.99999999999997,
            4.999465942382926,
            45.000686645507784,
            5.000152587890739,
        )
        assert tm.bbox_to_tiles((45, 5, 48, 6)) == (269425, 61166, 270882, 65535)
        assert tm.point_to_indices(45, 5) == (269425, 65535, 199, 255)
    except Exception as exc:
        assert False, f"'TileMatrixSet creation raises an exception: {exc}"



"""Provide functions to manipulate OGR / OSR entities
"""
# -- IMPORTS --

# standard library
import os
import re
from typing import Tuple

# 3rd party
from osgeo import gdal, ogr, osr

# package
from rok4.enums import ColorFormat

# -- GLOBALS --
ogr.UseExceptions()
osr.UseExceptions()
gdal.UseExceptions()

__SR_BOOK = {}


def srs_to_spatialreference(srs: str) -> "osgeo.osr.SpatialReference":
    """Convert coordinates system as string to OSR spatial reference

    Using a cache, to instanciate a Spatial Reference from a string only once.

    Args:
        srs (str): coordinates system PROJ4 compliant, with authority and code, like EPSG:3857 or IGNF:LAMB93

    Raises:
        RuntimeError: Provided SRS is invalid for OSR

    Returns:
        osgeo.osr.SpatialReference: Corresponding OSR spatial reference
    """

    global __SR_BOOK

    if srs.upper() not in __SR_BOOK:
        authority, code = srs.split(":", 1)

        sr = osr.SpatialReference()
        if authority.upper() == "EPSG":
            sr.ImportFromEPSG(int(code))
        else:
            sr.ImportFromProj4(f"+init={srs.upper()} +wktext")

        __SR_BOOK[srs.upper()] = sr

    return __SR_BOOK[srs.upper()]


def bbox_to_geometry(
    bbox: Tuple[float, float, float, float], densification: int = 0
) -> "osgeo.ogr.Geometry":
    """Convert bbox coordinates to OGR geometry

    Args:
        bbox (Tuple[float, float, float, float]): bounding box (xmin, ymin, xmax, ymax)
        densification (int, optional): Number of point to add for each side of bounding box. Defaults to 0.

    Raises:
        RuntimeError: Provided SRS is invalid for OSR

    Returns:
        osgeo.ogr.Geometry: Corresponding OGR geometry, with spatial reference if provided
    """

    ring = ogr.Geometry(ogr.wkbLinearRing)

    if densification > 0:
        step_x = (bbox[2] - bbox[0]) / (densification + 1)
        step_y = (bbox[3] - bbox[1]) / (densification + 1)

        for i in range(densification + 1):
            ring.AddPoint(bbox[0] + step_x * i, bbox[1])
        for i in range(densification + 1):
            ring.AddPoint(bbox[2], bbox[1] + step_y * i)
        for i in range(densification + 1):
            ring.AddPoint(bbox[2] - step_x * i, bbox[3])
        for i in range(densification + 1):
            ring.AddPoint(bbox[0], bbox[3] - step_y * i)
        ring.AddPoint(bbox[0], bbox[1])

    else:
        ring.AddPoint(bbox[0], bbox[1])
        ring.AddPoint(bbox[2], bbox[1])
        ring.AddPoint(bbox[2], bbox[3])
        ring.AddPoint(bbox[0], bbox[3])
        ring.AddPoint(bbox[0], bbox[1])

    geom = ogr.Geometry(ogr.wkbPolygon)
    geom.AddGeometry(ring)
    geom.SetCoordinateDimension(2)

    return geom


def reproject_bbox(
    bbox: Tuple[float, float, float, float], srs_src: str, srs_dst: str, densification: int = 5
) -> Tuple[float, float, float, float]:
    """Return bounding box in other coordinates system

    Points are added to be sure output bounding box contains input bounding box

    Args:
        bbox (Tuple[float, float, float, float]): bounding box (xmin, ymin, xmax, ymax) with source coordinates system
        srs_src (str): source coordinates system
        srs_dst (str): destination coordinates system
        densification (int, optional): Number of point to add for each side of bounding box. Defaults to 5.

    Returns:
        Tuple[float, float, float, float]: bounding box (xmin, ymin, xmax, ymax) with destination coordinates system
    """

    sr_src = srs_to_spatialreference(srs_src)
    sr_src_inv = sr_src.EPSGTreatsAsLatLong() or sr_src.EPSGTreatsAsNorthingEasting()

    sr_dst = srs_to_spatialreference(srs_dst)
    sr_dst_inv = sr_dst.EPSGTreatsAsLatLong() or sr_dst.EPSGTreatsAsNorthingEasting()

    if sr_src.IsSame(sr_dst) and sr_src_inv == sr_dst_inv:
        # Les système sont vraiment les même, avec le même ordre des axes
        return bbox
    elif sr_src.IsSame(sr_dst) and sr_src_inv != sr_dst_inv:
        # Les système sont les même pour OSR, mais l'ordre des axes est différent
        return (bbox[1], bbox[0], bbox[3], bbox[2])

    # Systèmes différents

    bbox_src = bbox_to_geometry(bbox, densification)
    bbox_src.AssignSpatialReference(sr_src)

    bbox_dst = bbox_src.Clone()
    os.environ["OGR_ENABLE_PARTIAL_REPROJECTION"] = "YES"
    bbox_dst.TransformTo(sr_dst)

    env = bbox_dst.GetEnvelope()
    return (env[0], env[2], env[1], env[3])


def reproject_point(
    point: Tuple[float, float],
    sr_src: "osgeo.osr.SpatialReference",
    sr_dst: "osgeo.osr.SpatialReference",
) -> Tuple[float, float]:
    """Reproject a point

    Args:
        point (Tuple[float, float]): source spatial reference point
        sr_src (osgeo.osr.SpatialReference): source spatial reference
        sr_dst (osgeo.osr.SpatialReference): destination spatial reference

    Returns:
        Tuple[float, float]: X/Y in destination spatial reference
    """

    sr_src_inv = sr_src.EPSGTreatsAsLatLong() or sr_src.EPSGTreatsAsNorthingEasting()
    sr_dst_inv = sr_dst.EPSGTreatsAsLatLong() or sr_dst.EPSGTreatsAsNorthingEasting()

    if sr_src.IsSame(sr_dst) and sr_src_inv == sr_dst_inv:
        # Les système sont vraiment les même, avec le même ordre des axes
        return (point[0], point[1])
    elif sr_src.IsSame(sr_dst) and sr_src_inv != sr_dst_inv:
        # Les système sont les même pour OSR, mais l'ordre des axes est différent
        return (point[1], point[0])

    # Systèmes différents
    ct = osr.CreateCoordinateTransformation(sr_src, sr_dst)
    x_dst, y_dst, z_dst = ct.TransformPoint(point[0], point[1])

    return (x_dst, y_dst)


def compute_bbox(source_dataset: gdal.Dataset) -> Tuple:
    """Image boundingbox computing method

    Args:
        source_dataset (gdal.Dataset): Dataset instanciated
          from the raster image

    Limitations:
        Image's axis must be parallel to SRS' axis

    Raises:
        AttributeError: source_dataset is not a gdal.Dataset instance.
        Exception: The dataset does not contain transform data.
    """
    bbox = None
    transform_vector = source_dataset.GetGeoTransform()
    if transform_vector is None:
        raise Exception(
            "No transform vector found in the dataset created from "
            + f"the following file : {source_dataset.GetFileList()[0]}"
        )
    width = source_dataset.RasterXSize
    height = source_dataset.RasterYSize
    x_range = (
        transform_vector[0],
        transform_vector[0] + width * transform_vector[1] + height * transform_vector[2],
    )
    y_range = (
        transform_vector[3],
        transform_vector[3] + width * transform_vector[4] + height * transform_vector[5],
    )
    spatial_ref = source_dataset.GetSpatialRef()
    if spatial_ref is not None and spatial_ref.GetDataAxisToSRSAxisMapping() == [2, 1]:
        # Coordonnées terrain de type (latitude, longitude)
        # => on permute les coordonnées terrain par rapport à l'image
        bbox = (min(y_range), min(x_range), max(y_range), max(x_range))
    else:
        # Coordonnées terrain de type (longitude, latitude) ou pas de SRS
        # => les coordonnées terrain sont dans le même ordre que celle de l'image
        bbox = (min(x_range), min(y_range), max(x_range), max(y_range))
    return bbox


def compute_format(dataset: gdal.Dataset, path: str = None) -> ColorFormat:
    """Image color format computing method

    Args:
        dataset (gdal.Dataset): Dataset instanciated from the image
        path (str, optionnal): path to the original file/object

    Raises:
        AttributeError: source_dataset is not a gdal.Dataset instance.
        Exception: No color band found or unsupported color format.
    """
    color_format = None
    if path is None:
        path = dataset.GetFileList()[0]
    if dataset.RasterCount < 1:
        raise Exception(f"Image {path} contains no color band.")

    band_1_datatype = dataset.GetRasterBand(1).DataType
    data_type_name = gdal.GetDataTypeName(band_1_datatype)
    data_type_size = gdal.GetDataTypeSize(band_1_datatype)
    color_interpretation = dataset.GetRasterBand(1).GetRasterColorInterpretation()
    color_name = None
    if color_interpretation is not None:
        color_name = gdal.GetColorInterpretationName(color_interpretation)
    compression_regex_match = re.search(r"COMPRESSION\s*=\s*PACKBITS", gdal.Info(dataset))

    if (
        data_type_name == "Byte"
        and data_type_size == 8
        and color_name == "Palette"
        and compression_regex_match
    ):
        # Compris par libTIFF comme du noir et blanc sur 1 bit
        color_format = ColorFormat.BIT
    elif data_type_name == "Byte" and data_type_size == 8:
        color_format = ColorFormat.UINT8
    elif data_type_name == "Float32" and data_type_size == 32:
        color_format = ColorFormat.FLOAT32
    else:
        raise Exception(
            f"Unsupported color format for image {path} : "
            + f"'{data_type_name}' ({data_type_size} bits)"
        )
    return color_format
